# Tests in this file work only with a working Mantid installation available in
# PYTHONPATH.
import importlib
import os
import sys
import tempfile
import unittest
import warnings

import numpy as np
import pytest
import scipp as sc

import scippneutron as scn
from scippneutron._utils import get_attrs

mantid = pytest.importorskip(
    'mantid.simpleapi', reason='Mantid framework is unavailable'
)


def memory_is_at_least_gb(required):
    import psutil

    total = psutil.virtual_memory().total / 1e9
    return total >= required


@pytest.mark.filterwarnings("ignore:convert_Workspace2D_to_data_array")
@pytest.mark.filterwarnings("ignore:convert_EventWorkspace_to_data_array")
@pytest.mark.filterwarnings("ignore:convert_MDHistoWorkspace_to_data_array")
@pytest.mark.filterwarnings("ignore:scippneutron.array_from_mantid")
@pytest.mark.skipif(not memory_is_at_least_gb(4), reason='Insufficient virtual memory')
class TestMantidConversion(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # This is from the Mantid system-test data
        filename = "CNCS_51936_event.nxs"
        # This needs OutputWorkspace specified, as it doesn't
        # pick up the name from the class variable name
        cls.base_event_ws = mantid.LoadEventNexus(
            scn.data.get_path(filename),
            OutputWorkspace="test_ws{}".format(__file__),
            SpectrumMax=200,
            StoreInADS=False,
        )

    def test_Workspace2D_data_array(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000, PreserveEvents=False)
        d = scn.mantid.convert_Workspace2D_to_data_array(ws)
        assert get_attrs(d)["run_start"].value == "2012-05-21T15:14:56.279289666"
        assert d.data.unit == sc.units.counts
        for i in range(ws.getNumberHistograms()):
            assert np.all(np.equal(d.values[i], ws.readY(i)))
            assert np.all(np.equal(d.variances[i], ws.readE(i) * ws.readE(i)))
        assert d.coords['spectrum'].dtype == sc.DType.int32
        assert d.coords['spectrum'].unit is None
        assert d.coords['tof'].dtype == sc.DType.float64

    def test_Workspace2D(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000, PreserveEvents=False)
        d = scn.mantid.convert_Workspace2D_to_data_group(ws)
        assert d["run_start"].value == "2012-05-21T15:14:56.279289666"
        assert d["data"].unit == sc.units.counts
        for i in range(ws.getNumberHistograms()):
            assert np.all(np.equal(d["data"].values[i], ws.readY(i)))
            assert np.all(np.equal(d["data"].variances[i], ws.readE(i) * ws.readE(i)))
        assert d["data"].coords['spectrum'].dtype == sc.DType.int32
        assert d["data"].coords['spectrum'].unit is None
        assert d["data"].coords['tof'].dtype == sc.DType.float64

    def test_EventWorkspace_data_array(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000)

        binned_mantid = scn.mantid.convert_Workspace2D_to_data_array(ws)

        target_tof = binned_mantid.coords['tof']
        d = scn.mantid.convert_EventWorkspace_to_data_array(
            eventWS, load_pulse_times=False
        )
        histogrammed = d.hist(tof=target_tof)

        delta = sc.sum(binned_mantid - histogrammed, 'spectrum')
        delta = sc.sum(delta, 'tof')
        self.assertLess(np.abs(delta.value), 1e-5)

    def test_EventWorkspace(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000)

        binned_mantid = scn.mantid.convert_Workspace2D_to_data_group(ws)["data"]

        target_tof = binned_mantid.coords['tof']
        d = scn.mantid.convert_EventWorkspace_to_data_group(
            eventWS, load_pulse_times=False
        )
        assert d["run_start"].value == "2012-05-21T15:14:56.279289666"

        histogrammed = d["data"].hist(tof=target_tof)
        delta = sc.sum(binned_mantid - histogrammed, 'spectrum')
        delta = sc.sum(delta, 'tof')
        assert np.abs(delta.value) < 1e-5

    def test_EventWorkspace_empty_event_list_consistent_bin_indices(self):
        ws = mantid.CloneWorkspace(self.base_event_ws)
        ws.getSpectrum(ws.getNumberHistograms() - 1).clear(removeDetIDs=True)

        dg = scn.mantid.convert_EventWorkspace_to_data_group(ws, load_pulse_times=False)
        assert dg["run_start"].value == "2012-05-21T15:14:56.279289666"
        da = dg["data"]
        assert da.bins.size()['spectrum', -1]['tof', 0].value == 0

    def test_comparison(self):
        a = scn.mantid.convert_EventWorkspace_to_data_group(
            self.base_event_ws, load_pulse_times=False
        )
        b = a.copy()
        assert sc.identical(a, b)

    def test_advanced_geometry(self):
        # basic test that positions are approximately equal for detectors for
        # CNCS given advanced and basic geometry calculation routes
        x = scn.from_mantid(self.base_event_ws, advanced_geometry=False)
        y = scn.from_mantid(self.base_event_ws, advanced_geometry=True)
        assert sc.allclose(x["data"].coords['position'], y["data"].coords['position'])

    def test_advanced_geometry_with_absent_shape(self):
        # single bank 3 by 3
        ws = mantid.CreateSampleWorkspace(
            NumBanks=1, BankPixelWidth=3, StoreInADS=False
        )
        # Save and reload trick to purge sample shape info
        file_name = "example_geometry.nxs"
        geom_path = os.path.join(tempfile.gettempdir(), file_name)
        mantid.SaveNexusGeometry(ws, geom_path)  # Does not save shape info
        assert os.path.isfile(geom_path)  # sanity check
        out = mantid.LoadEmptyInstrument(
            Filename=geom_path, StoreInADS=False
        )  # reload without sample info
        os.remove(geom_path)

        assert not out.componentInfo().hasValidShape(0)  # sanity check
        dg = scn.mantid.from_mantid(out, advanced_geometry=True)
        # Shapes have zero size
        assert sc.identical(
            sc.sum(dg['shape']), sc.vector(value=[0, 0, 0], unit=sc.units.m)
        )

    def test_advanced_geometry_detector_info(self):
        dg = scn.from_mantid(self.base_event_ws, advanced_geometry=True)
        detector_info = dg['data'].coords['detector_info'].value
        assert detector_info.dim == 'detector'
        assert detector_info.coords['detector'].unit is None
        assert detector_info.coords['spectrum'].unit is None

    def test_EventWorkspace_no_y_unit(self):
        tiny_event_ws = mantid.CreateSampleWorkspace(
            WorkspaceType='Event', NumBanks=1, NumEvents=1
        )
        d = scn.mantid.convert_EventWorkspace_to_data_array(
            tiny_event_ws, load_pulse_times=False
        )
        assert d.data.bins.constituents['data'].unit == sc.units.counts
        tiny_event_ws.setYUnit('')
        d = scn.mantid.convert_EventWorkspace_to_data_array(
            tiny_event_ws, load_pulse_times=False
        )
        assert d.data.bins.constituents['data'].unit == sc.units.one

    def test_from_mantid_LoadEmptyInstrument(self):
        ws = mantid.LoadEmptyInstrument(InstrumentName='PG3')
        scn.from_mantid(ws)

    def test_from_mantid_CreateWorkspace(self):
        dataX = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
        dataY = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
        ws = mantid.CreateWorkspace(
            DataX=dataX, DataY=dataY, NSpec=4, UnitX="Wavelength"
        )
        d = scn.from_mantid(ws)
        assert d['data'].unit == sc.units.dimensionless

    def test_unit_conversion(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000, PreserveEvents=False)
        tmp = scn.mantid.convert_Workspace2D_to_data_array(ws)
        target_tof = tmp.coords['tof']
        ws = mantid.ConvertUnits(
            InputWorkspace=ws, Target="Wavelength", EMode="Elastic"
        )
        converted_mantid = scn.mantid.convert_Workspace2D_to_data_group(ws)

        da = scn.mantid.convert_EventWorkspace_to_data_array(
            eventWS, load_pulse_times=False
        )
        da = da.hist(tof=target_tof)
        d = sc.Dataset(data={da.name: da})
        converted = scn.convert(d, 'tof', 'wavelength', scatter=True)

        assert sc.allclose(
            converted_mantid['data'].data, converted[""].data.to(dtype='float64')
        )
        assert sc.allclose(
            converted_mantid['data'].coords['wavelength'],
            converted.coords['wavelength'],
        )

    def test_inelastic_unit_conversion(self):
        eventWS = self.base_event_ws
        ws_deltaE = mantid.ConvertUnits(
            eventWS, Target='DeltaE', EMode='Direct', EFixed=3
        )
        ref = scn.from_mantid(ws_deltaE)['data']
        da = scn.from_mantid(eventWS)['data']
        # Boost and Mantid use CODATA 2006. This test passes if we manually
        # change the implementation to use the old constants. Alternatively
        # we can correct for this by scaling L1^2 or L2^2, and this was also
        # confirmed in C++. Unfortunately only positions are accessible to
        # correct for this here, and due to precision issues with
        # dot/norm/sqrt this doesn't actually fix the test. We additionally
        # exclude low TOF region, and bump relative and absolute accepted
        # errors from 1e-8 to 1e-5.
        m_n_2006 = 1.674927211
        m_n_2018 = 1.67492749804
        e_2006 = 1.602176487
        e_2018 = 1.602176634
        scale = (m_n_2006 / m_n_2018) / (e_2006 / e_2018)
        da.coords['source_position'] *= np.sqrt(scale)
        da.coords['position'] *= np.sqrt(scale)
        low_tof = da.bins.constituents['data'].coords['tof'] < 49000.0 * sc.units.us
        da.coords['incident_energy'] = 3.0 * sc.units.meV
        da = scn.convert(da, 'tof', 'energy_transfer', scatter=True)
        assert sc.all(
            sc.isnan(da.coords['energy_transfer'])
            | sc.isclose(
                da.coords['energy_transfer'],
                ref.coords['energy_transfer'],
                atol=1e-8 * sc.units.meV,
                rtol=1e-8 * sc.units.one,
            )
        ).value
        assert sc.all(
            low_tof
            | sc.isnan(da.bins.constituents['data'].coords['energy_transfer'])
            | sc.isclose(
                da.bins.constituents['data'].coords['energy_transfer'],
                ref.bins.constituents['data'].coords['energy_transfer'],
                atol=1e-5 * sc.units.meV,
                rtol=1e-5 * sc.units.one,
            )
        ).value

    @staticmethod
    def _mask_bins_and_spectra(ws, xmin, xmax, num_spectra, indices=None):
        masked_ws = mantid.MaskBins(
            ws, XMin=xmin, XMax=xmax, InputWorkspaceIndexSet=indices
        )

        # mask the first 3 spectra
        for i in range(num_spectra):
            masked_ws.spectrumInfo().setMasked(i, True)

        return masked_ws

    def test_Workspace2D_common_bins_masks(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000, PreserveEvents=False)
        ws_x = ws.readX(0)

        # mask the first 3 bins, range is taken as [XMin, XMax)
        masked_ws = self._mask_bins_and_spectra(
            ws, xmin=ws_x[0], xmax=ws_x[3], num_spectra=3
        )

        assert masked_ws.isCommonBins()

        da = scn.mantid.convert_Workspace2D_to_data_group(masked_ws)['data']

        np.testing.assert_array_equal(da.masks["bin"].values[0:3], [True, True, True])

        np.testing.assert_array_equal(
            da.masks["spectrum"].values[0:3], [True, True, True]
        )

    def test_Workspace2D_common_bins_not_common_masks(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000, PreserveEvents=False)
        ws_x = ws.readX(0)

        # mask first 3 bins in first 3 spectra, range is taken as [XMin, XMax)
        masked_ws = self._mask_bins_and_spectra(
            ws, xmin=ws_x[0], xmax=ws_x[3], num_spectra=3, indices='0-2'
        )

        assert masked_ws.isCommonBins()

        da = scn.mantid.convert_Workspace2D_to_data_group(masked_ws)['data']

        mask = sc.zeros(dims=da.dims, shape=da.shape, dtype=sc.DType.bool)
        mask['spectrum', 0:3]['tof', 0:3] |= sc.scalar(True)
        assert sc.identical(da.masks['bin'], mask)

        np.testing.assert_array_equal(
            da.masks["spectrum"].values[0:3], [True, True, True]
        )

    def test_Workspace2D_not_common_bins_masks(self):
        eventWS = self.base_event_ws
        ws = mantid.Rebin(eventWS, 10000, PreserveEvents=False)
        ws = mantid.ConvertUnits(ws, "Wavelength", EMode="Direct", EFixed=0.1231)

        # these X values will mask different number of bins
        masked_ws = self._mask_bins_and_spectra(
            ws, -214, -192, num_spectra=3, indices='0-40'
        )

        assert not masked_ws.isCommonBins()

        da = scn.mantid.convert_Workspace2D_to_data_group(masked_ws)['data']

        # bin with 3 masks
        np.testing.assert_array_equal(da.masks["bin"].values[0], [True, True, False])

        # bin with only 2
        np.testing.assert_array_equal(da.masks["bin"].values[31], [True, True, False])

        np.testing.assert_array_equal(
            da.masks["spectrum"].values[0:3], [True, True, True]
        )

    @staticmethod
    def check_monitor_metadata(monitor):
        assert 'position' in monitor.coords
        assert 'source_position' in monitor.coords
        assert not monitor.coords['sample_position'].aligned
        # Absence of the following is not crucial, but currently there is
        # no need for these, and it avoids duplication:
        assert 'detector_info' not in monitor.coords
        assert 'sample' not in monitor.coords
        assert (
            'SampleTemp' not in monitor.coords
        ), "Expect run logs not be duplicated in monitor workspaces"

    @staticmethod
    def check_monitor_metadata_old(monitor):
        # TODO remove once load and array_from_mantid are removed
        assert 'position' in monitor.coords
        assert 'source_position' in monitor.coords
        assert 'sample_position' not in monitor.coords
        assert 'sample_position' in get_attrs(monitor)
        # Absence of the following is not crucial, but currently there is
        # no need for these, and it avoids duplication:
        assert 'detector_info' not in monitor.coords
        assert 'sample' not in monitor.coords
        assert (
            'SampleTemp' not in monitor.coords
        ), "Expect run logs not be duplicated in monitor workspaces"

    def test_Workspace2D_with_separate_monitors(self):
        from mantid.simpleapi import mtd

        mtd.clear()
        # This test would use 20 GB of memory if "SpectrumMax" was not set
        dg = scn.load_with_mantid(
            scn.data.get_path("WISH00016748.raw"),
            mantid_args={"LoadMonitors": "Separate", "SpectrumMax": 10000},
        )
        assert len(mtd) == 0, mtd.getObjectNames()
        expected_monitor_names = {
            "monitor1",
            "monitor2",
            "monitor3",
            "monitor4",
            "monitor5",
        }
        assert dg["monitors"].keys() == expected_monitor_names

        for monitor in dg["monitors"].values():
            assert isinstance(monitor, sc.DataGroup)
            assert monitor.shape == (4471,)
            self.check_monitor_metadata(monitor['data'])

    @pytest.mark.filterwarnings("ignore:scippneutron.load")
    def test_load_Workspace2D_with_separate_monitors(self):
        from mantid.simpleapi import mtd

        mtd.clear()
        # This test would use 20 GB of memory if "SpectrumMax" was not set
        da = scn.load(
            scn.data.get_path("WISH00016748.raw"),
            mantid_args={"LoadMonitors": "Separate", "SpectrumMax": 10000},
        )
        assert len(mtd) == 0, mtd.getObjectNames()
        expected_monitor_names = {
            "monitor1",
            "monitor2",
            "monitor3",
            "monitor4",
            "monitor5",
        }
        assert expected_monitor_names.issubset(get_attrs(da).keys())

        for monitor_name in expected_monitor_names:
            monitor = get_attrs(da)[monitor_name].value
            assert isinstance(monitor, sc.DataArray)
            assert monitor.shape == (4471,)
            self.check_monitor_metadata_old(monitor)

    def test_Workspace2D_with_include_monitors(self):
        from mantid.simpleapi import mtd

        mtd.clear()
        # This test would use 20 GB of memory if "SpectrumMax" was not set
        dg = scn.load_with_mantid(
            scn.data.get_path("WISH00016748.raw"),
            mantid_args={"LoadMonitors": "Include", "SpectrumMax": 100},
        )
        assert len(mtd) == 0, mtd.getObjectNames()
        expected_monitor_names = {
            "monitor1",
            "monitor2",
            "monitor3",
            "monitor4",
            "monitor5",
        }
        assert dg["monitors"].keys() == expected_monitor_names
        for monitor in dg["monitors"].values():
            assert isinstance(monitor, sc.DataGroup)
            assert monitor.shape == (4471,)
            self.check_monitor_metadata(monitor['data'])

    @pytest.mark.filterwarnings("ignore:scippneutron.load")
    def test_load_Workspace2D_with_include_monitors(self):
        from mantid.simpleapi import mtd

        mtd.clear()
        # This test would use 20 GB of memory if "SpectrumMax" was not set
        da = scn.load(
            scn.data.get_path("WISH00016748.raw"),
            mantid_args={"LoadMonitors": "Include", "SpectrumMax": 100},
        )
        assert len(mtd) == 0, mtd.getObjectNames()
        expected_monitor_names = {
            "monitor1",
            "monitor2",
            "monitor3",
            "monitor4",
            "monitor5",
        }
        assert expected_monitor_names.issubset(get_attrs(da).keys())
        for monitor_name in expected_monitor_names:
            monitor = get_attrs(da)[monitor_name].value
            assert isinstance(monitor, sc.DataArray)
            assert monitor.shape == (4471,)
            self.check_monitor_metadata_old(monitor)

    def test_EventWorkspace_with_monitors(self):
        from mantid.simpleapi import mtd

        mtd.clear()
        dg = scn.load_with_mantid(
            scn.data.get_path("CNCS_51936_event.nxs"),
            mantid_args={"LoadMonitors": True, "SpectrumMax": 1},
        )
        assert len(mtd) == 0, mtd.getObjectNames()
        expected_monitor_attrs = {"monitor2", "monitor3"}
        assert dg["monitors"].keys() == expected_monitor_attrs
        for monitor in dg["monitors"].values():
            assert isinstance(monitor, sc.DataGroup)
            assert monitor.shape == (200001,)
            self.check_monitor_metadata(monitor['data'])

    @pytest.mark.filterwarnings("ignore:scippneutron.load")
    def test_load_EventWorkspace_with_monitors(self):
        from mantid.simpleapi import mtd

        mtd.clear()
        da = scn.load(
            scn.data.get_path("CNCS_51936_event.nxs"),
            mantid_args={"LoadMonitors": True, "SpectrumMax": 1},
        )
        self.assertEqual(len(mtd), 0, mtd.getObjectNames())
        attrs = [str(key) for key in get_attrs(da).keys()]
        expected_monitor_attrs = {"monitor2", "monitor3"}
        assert expected_monitor_attrs.issubset(attrs)
        for monitor_name in expected_monitor_attrs:
            monitor = get_attrs(da)[monitor_name].value
            assert isinstance(monitor, sc.DataArray)
            assert monitor.shape == (200001,)
            self.check_monitor_metadata_old(monitor)

    def test_mdhisto_workspace_q(self):
        from mantid.simpleapi import BinMD, CreateMDWorkspace, FakeMDEventData

        md_event = CreateMDWorkspace(
            Dimensions=3,
            Extents=[-10, 10, -10, 10, -10, 10],
            Names='Q_x,Q_y,Q_z',
            Units='U,U,U',
            Frames='QLab,QLab,QLab',
            StoreInADS=False,
        )
        FakeMDEventData(
            InputWorkspace=md_event, PeakParams=[100000, 0, 0, 0, 1], StoreInADS=False
        )  # Add Peak
        md_histo = BinMD(
            InputWorkspace=md_event,
            AlignedDim0='Q_y,-10,10,3',
            AlignedDim1='Q_x,-10,10,4',
            AlignedDim2='Q_z,-10,10,5',
            StoreInADS=False,
        )

        histo_data_group = scn.mantid.convert_MDHistoWorkspace_to_data_group(md_histo)
        histo_data_array = histo_data_group['data']

        assert histo_data_array.coords['Q_x'].shape == (4,)
        assert histo_data_array.coords['Q_y'].shape == (3,)
        assert histo_data_array.coords['Q_z'].shape == (5,)
        assert (
            histo_data_array.coords['Q_x'].unit
            == sc.units.dimensionless / sc.units.angstrom
        )
        assert (
            histo_data_array.coords['Q_y'].unit
            == sc.units.dimensionless / sc.units.angstrom
        )
        assert (
            histo_data_array.coords['Q_z'].unit
            == sc.units.dimensionless / sc.units.angstrom
        )
        assert histo_data_array.shape == (3, 4, 5)

        # Sum over 2 dimensions to simplify finding max.
        max_1d = sc.sum(sc.sum(histo_data_array, dim='Q_y'), dim='Q_x').values
        max_index = np.argmax(max_1d)
        # Check position of max 'peak'
        assert np.floor(len(max_1d) / 2) == max_index
        # All events in central 'peak'
        assert 100000 == max_1d[max_index]

        assert 'nevents' in histo_data_group

    def test_mdhisto_workspace_q_data_array(self):
        from mantid.simpleapi import BinMD, CreateMDWorkspace, FakeMDEventData

        md_event = CreateMDWorkspace(
            Dimensions=3,
            Extents=[-10, 10, -10, 10, -10, 10],
            Names='Q_x,Q_y,Q_z',
            Units='U,U,U',
            Frames='QLab,QLab,QLab',
            StoreInADS=False,
        )
        FakeMDEventData(
            InputWorkspace=md_event, PeakParams=[100000, 0, 0, 0, 1], StoreInADS=False
        )  # Add Peak
        md_histo = BinMD(
            InputWorkspace=md_event,
            AlignedDim0='Q_y,-10,10,3',
            AlignedDim1='Q_x,-10,10,4',
            AlignedDim2='Q_z,-10,10,5',
            StoreInADS=False,
        )

        histo_data_array = scn.mantid.convert_MDHistoWorkspace_to_data_array(md_histo)

        self.assertEqual(histo_data_array.coords['Q_x'].shape, (4,))
        self.assertEqual(histo_data_array.coords['Q_y'].shape, (3,))
        self.assertEqual(histo_data_array.coords['Q_z'].shape, (5,))
        self.assertEqual(
            histo_data_array.coords['Q_x'].unit,
            sc.units.dimensionless / sc.units.angstrom,
        )
        self.assertEqual(
            histo_data_array.coords['Q_y'].unit,
            sc.units.dimensionless / sc.units.angstrom,
        )
        self.assertEqual(
            histo_data_array.coords['Q_z'].unit,
            sc.units.dimensionless / sc.units.angstrom,
        )

        self.assertEqual(histo_data_array.shape, (3, 4, 5))

        # Sum over 2 dimensions to simplify finding max.
        max_1d = sc.sum(sc.sum(histo_data_array, dim='Q_y'), dim='Q_x').values
        max_index = np.argmax(max_1d)
        # Check position of max 'peak'
        self.assertEqual(np.floor(len(max_1d) / 2), max_index)
        # All events in central 'peak'
        self.assertEqual(100000, max_1d[max_index])

        self.assertTrue('nevents' in get_attrs(histo_data_array))

    def test_mdhisto_workspace_many_dims(self):
        from mantid.simpleapi import BinMD, CreateMDWorkspace, FakeMDEventData

        md_event = CreateMDWorkspace(
            Dimensions=4,
            Extents=[-10, 10, -10, 10, -10, 10, -10, 10],
            Names='deltae,y,z,T',
            Units='U,U,U,U',
            StoreInADS=False,
        )
        FakeMDEventData(
            InputWorkspace=md_event,
            PeakParams=[100000, 0, 0, 0, 0, 1],
            StoreInADS=False,
        )  # Add Peak
        md_histo = BinMD(
            InputWorkspace=md_event,
            AlignedDim0='deltae,-10,10,3',
            AlignedDim1='y,-10,10,4',
            AlignedDim2='z,-10,10,5',
            AlignedDim3='T,-10,10,7',
            StoreInADS=False,
        )

        histo_data_group = scn.mantid.convert_MDHistoWorkspace_to_data_group(md_histo)
        assert len(histo_data_group.dims) == 4

    def test_mdhisto_workspace_many_dims_array(self):
        from mantid.simpleapi import BinMD, CreateMDWorkspace, FakeMDEventData

        md_event = CreateMDWorkspace(
            Dimensions=4,
            Extents=[-10, 10, -10, 10, -10, 10, -10, 10],
            Names='deltae,y,z,T',
            Units='U,U,U,U',
            StoreInADS=False,
        )
        FakeMDEventData(
            InputWorkspace=md_event,
            PeakParams=[100000, 0, 0, 0, 0, 1],
            StoreInADS=False,
        )  # Add Peak
        md_histo = BinMD(
            InputWorkspace=md_event,
            AlignedDim0='deltae,-10,10,3',
            AlignedDim1='y,-10,10,4',
            AlignedDim2='z,-10,10,5',
            AlignedDim3='T,-10,10,7',
            StoreInADS=False,
        )

        histo_data_array = scn.mantid.convert_MDHistoWorkspace_to_data_array(md_histo)
        self.assertEqual(4, len(histo_data_array.dims))

    def test_to_workspace_2d_no_error(self):
        from mantid.simpleapi import mtd

        mtd.clear()

        # All Dims for which support is expected are
        # tested in the parametrized test.
        # Just set this one to a working one to avoid
        # generating many repetitive tests.
        param_dim = 'tof'
        data_len = 2
        expected_bins = data_len + 1
        expected_number_spectra = 10

        y = sc.Variable(
            dims=['spectrum', param_dim],
            values=np.random.rand(expected_number_spectra, data_len),
        )

        x = sc.Variable(
            dims=['spectrum', param_dim],
            values=np.arange(
                expected_number_spectra * expected_bins, dtype=np.float64
            ).reshape((expected_number_spectra, expected_bins)),
        )
        data = sc.DataArray(data=y, coords={param_dim: x})

        ws = scn.to_mantid(data, param_dim)

        assert len(ws.readX(0)) == expected_bins
        assert ws.getNumberHistograms() == expected_number_spectra
        # check that no workspaces have been leaked in the ADS
        assert len(mtd) == 0, f"Workspaces present: {mtd.getObjectNames()}"

        for i in range(expected_number_spectra):
            np.testing.assert_array_equal(ws.readX(i), x['spectrum', i].values)
            np.testing.assert_array_equal(ws.readY(i), y['spectrum', i].values)
            np.testing.assert_array_equal(ws.readE(i), np.sqrt(y['spectrum', i].values))

    def test_fit(self):
        """
        Tests that the fit executes, and the outputs
        are moved into the dataset. Does not check the fit values.
        """
        from mantid.simpleapi import mtd

        mtd.clear()

        data = scn.load_with_mantid(scn.data.get_path("iris26176_graphite002_sqw.nxs"))

        params, diff = scn.fit(
            data['Q', 0]['data'],
            mantid_args={
                'Function': 'name=LinearBackground,A0=0,A1=1',
                'StartX': 0,
                'EndX': 3,
            },
        )

        # check that no workspaces have been leaked in the ADS
        assert len(mtd) == 0
        assert 'data' in diff
        assert 'calculated' in diff
        assert 'diff' in diff
        assert 'status' in params.coords
        assert 'function' in params.coords
        assert 'cost_function' in params.coords
        assert 'chi^2/d.o.f.' in params.coords

    def test_convert_array_run_log_to_attrs(self):
        # Given a Mantid workspace with a run log
        target = mantid.CloneWorkspace(self.base_event_ws)
        log_name = "SampleTemp"
        self.assertTrue(
            target.run().hasProperty(log_name),
            f"Expected input workspace to have a {log_name} run log",
        )

        # When the workspace is converted to a scipp data array
        d = scn.mantid.convert_EventWorkspace_to_data_array(target, False)

        # Then the data array contains the run log as an unaligned coord
        self.assertTrue(
            np.allclose(
                target.run()[log_name].value,
                get_attrs(d)[log_name].values.data.values,
            ),
            "Expected values in the unaligned coord to match "
            "the original run log from the Mantid workspace",
        )
        self.assertEqual(get_attrs(d)[log_name].values.unit, sc.units.K)
        self.assertTrue(
            np.array_equal(
                target.run()[log_name].times.astype('datetime64[ns]'),
                get_attrs(d)[log_name].values.coords["time"].values,
            ),
            "Expected times in the unaligned coord to match "
            "the original run log from the Mantid workspace",
        )

    def test_convert_scalar_run_log_to_attrs(self):
        # Given a Mantid workspace with a run log
        target = mantid.CloneWorkspace(self.base_event_ws)
        log_name = "start_time"
        self.assertTrue(
            target.run().hasProperty(log_name),
            f"Expected input workspace to have a {log_name} run log",
        )

        # When the workspace is converted to a scipp data array
        d = scn.mantid.convert_EventWorkspace_to_data_array(target, False)

        # Then the data array contains the run log as an unaligned coord
        self.assertEqual(
            target.run()[log_name].value,
            get_attrs(d)[log_name].value,
            "Expected value of the unaligned coord to match "
            "the original run log from the Mantid workspace",
        )

    def test_warning_raised_when_convert_run_log_with_unrecognised_units(self):
        target = mantid.CloneWorkspace(self.base_event_ws)
        target.getRun()['LambdaRequest'].units = 'abcde'
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")
            scn.mantid.convert_EventWorkspace_to_data_array(target, False)
            assert len(caught_warnings) > 0, (
                "Expected warnings due to some run logs "
                "having unrecognised units strings"
            )
            assert any(
                "unrecognised units" in str(caught_warning.message)
                for caught_warning in caught_warnings
            )

    def test_no_warning_raised_explicitly_dimensionless_run_log(self):
        target = mantid.CloneWorkspace(self.base_event_ws)
        with warnings.catch_warnings(record=True) as caught_warnings:
            scn.mantid.convert_EventWorkspace_to_data_array(target, False)
            original_number_of_warnings = len(caught_warnings)

        # Add an explicitly dimensionless log
        mantid.AddSampleLog(
            Workspace=target,
            LogName='dimensionless_log',
            LogText='1',
            LogType='Number',
            LogUnit='dimensionless',
        )

        with warnings.catch_warnings(record=True) as caught_warnings:
            scn.mantid.convert_EventWorkspace_to_data_array(target, False)
            assert len(caught_warnings) == original_number_of_warnings, (
                "Expected no extra warning about unrecognised units "
                "from explicitly dimensionless log"
            )

    def test_set_sample(self):
        target = mantid.CloneWorkspace(self.base_event_ws)
        d = scn.mantid.convert_EventWorkspace_to_data_array(target, False)
        get_attrs(d)["sample"].value.setThickness(3)
        # before
        self.assertNotEqual(3, target.sample().getThickness())
        target.setSample(get_attrs(d)["sample"].value)
        # after
        self.assertEqual(3, target.sample().getThickness())

    def test_sample_ub(self):
        ws = mantid.CreateWorkspace(DataY=np.ones(1), DataX=np.arange(2))
        args = {'a': 1, 'b': 1, 'c': 1, 'alpha': 90, 'beta': 90, 'gamma': 90}
        mantid.SetUB(ws, **args)
        d = scn.mantid.from_mantid(ws)
        assert sc.identical(
            d['sample_ub'],
            sc.spatial.linear_transform(
                value=ws.sample().getOrientedLattice().getUB(),
                unit=sc.units.angstrom**-1,
            ),
        )
        assert sc.identical(
            d['sample_u'],
            sc.spatial.linear_transform(value=ws.sample().getOrientedLattice().getU()),
        )

    def test_sample_without_ub(self):
        ws = mantid.CreateWorkspace(DataY=np.ones(1), DataX=np.arange(2))
        assert not ws.sample().hasOrientedLattice()  # Sanity check input
        d = scn.mantid.from_mantid(ws)
        assert "sample_ub" not in d
        assert "sample_u" not in d

    def _exec_to_spherical(self, x, y, z):
        in_out = sc.Dataset(
            {
                'x': sc.scalar(x, unit=sc.units.m),
                'y': sc.scalar(y, unit=sc.units.m),
                'z': sc.scalar(z, unit=sc.units.m),
            }
        )
        point = sc.spatial.as_vectors(
            in_out['x'].data, in_out['y'].data, in_out['z'].data
        )
        scn.mantid._to_spherical(point, in_out)
        return in_out

    def test_spherical_conversion(self):
        x = 1.0
        y = 1.0
        z = 0.0
        spherical = self._exec_to_spherical(x, y, z)
        assert spherical['r'].value == np.sqrt(x**2 + y**2 + z**2)
        assert spherical['t'].value == np.arccos(z / np.sqrt(x**2 + y**2 + z**2))
        # Phi now should be between 0 and pi
        assert spherical['p-delta'].value == (3.0 / 4) * np.pi
        assert spherical['p-sign'].value > 0.0
        x = -1.0
        spherical = self._exec_to_spherical(x, y, z)
        assert spherical['p-delta'].value == (1.0 / 4) * np.pi
        assert spherical['p-sign'].value > 0.0
        # Phi now should be between 0 and -pi
        y = -1.0
        spherical = self._exec_to_spherical(x, y, z)
        assert spherical['p-delta'].value == (1.0 / 4) * np.pi
        assert spherical['p-sign'].value < 0.0

    def test_detector_positions(self):
        from mantid.kernel import V3D

        eventWS = mantid.CloneWorkspace(self.base_event_ws)
        comp_info = eventWS.componentInfo()
        small_offset = V3D(0.01, 0.01, 0.01)
        comp_info.setPosition(
            comp_info.source(), comp_info.samplePosition() + small_offset
        )
        moved = scn.mantid.convert_Workspace2D_to_data_array(eventWS)
        moved_det_position = moved.coords["position"]
        unmoved = scn.mantid.convert_Workspace2D_to_data_array(eventWS)
        unmoved_det_positions = unmoved.coords["position"]
        # Moving the sample accounted for in position calculations
        # but should not yield change to final detector positions
        self.assertTrue(
            np.all(np.isclose(moved_det_position.values, unmoved_det_positions.values))
        )

    def test_validate_units(self):
        acceptable = ["wavelength", "Wavelength"]
        for i in acceptable:
            ret = scn.mantid.validate_dim_and_get_mantid_string(i)
            self.assertEqual(ret, "Wavelength")

    def test_validate_units_throws(self):
        not_acceptable = [None, "None", "wavlength", 1, 1.0, ["wavelength"]]
        for i in not_acceptable:
            with self.assertRaises(RuntimeError):
                scn.mantid.validate_dim_and_get_mantid_string(i)

    def test_WorkspaceGroup_parsed_correctly(self):
        from mantid.simpleapi import CreateSampleWorkspace, GroupWorkspaces, mtd

        CreateSampleWorkspace(OutputWorkspace="ws1")
        CreateSampleWorkspace(OutputWorkspace="ws2")
        CreateSampleWorkspace(OutputWorkspace="ws3")
        GroupWorkspaces(InputWorkspaces="ws1,ws2,ws3", OutputWorkspace="NewGroup")

        converted_group = scn.from_mantid(mtd["NewGroup"])
        converted_single = scn.from_mantid(mtd["ws1"])

        assert len(converted_group) == 3
        assert sc.identical(converted_group['ws1'], converted_single)

        mtd.clear()


@pytest.mark.skipif(not memory_is_at_least_gb(8), reason='Insufficient virtual memory')
def test_load_mcstas_data():
    wsg = mantid.LoadMcStas(
        scn.data.get_path('mcstas_sans.h5'), OutputWorkspace="test_mcstas_sans_wsg"
    )
    ws = wsg[list(wsg.getNames()).index('EventData_test_mcstas_sans_wsg')]
    dg = scn.from_mantid(ws)

    for i in range(ws.getNumberHistograms()):
        np.testing.assert_array_equal(dg['data'].coords['tof'].values, ws.readX(i))
        spec = ws.getSpectrum(i)
        da_spec = dg['data']['tof', 0]['spectrum', i]
        bin_sizes = da_spec.bins.size()
        assert spec.getNumberEvents() == bin_sizes.value

        np.testing.assert_array_equal(
            spec.getTofs(), da_spec.bins.coords['tof'].values.values
        )
        np.testing.assert_array_equal(
            spec.getPulseTimesAsNumpy(), da_spec.bins.coords['pulse_time'].value.values
        )
        np.testing.assert_array_equal(spec.getWeights(), da_spec.bins.data.value.values)


def test_to_rot_from_vectors():
    a = sc.vector(value=[1, 0, 0])
    b = sc.vector(value=[0, 1, 0])
    rot = scn.mantid._rot_from_vectors(a, b)
    assert np.allclose((rot * a).value, b.value)
    rot = scn.mantid._rot_from_vectors(b, a)
    assert np.allclose((rot * b).value, a.value)


@pytest.mark.skipif(not memory_is_at_least_gb(8), reason='Insufficient virtual memory')
@pytest.mark.parametrize(
    "param_dim",
    ('tof', 'wavelength', 'energy', 'dspacing', 'Q', 'Q^2', 'energy_transfer'),
)
def test_to_workspace_2d(param_dim):
    from mantid.simpleapi import mtd

    mtd.clear()

    data_len = 2
    expected_bins = data_len + 1
    expected_number_spectra = 10

    y = sc.Variable(
        dims=['spectrum', param_dim],
        values=np.random.rand(expected_number_spectra, data_len),
        variances=np.random.rand(expected_number_spectra, data_len),
    )

    x = sc.Variable(
        dims=['spectrum', param_dim],
        values=np.arange(
            expected_number_spectra * expected_bins, dtype=np.float64
        ).reshape((expected_number_spectra, expected_bins)),
    )
    data = sc.DataArray(data=y, coords={param_dim: x})

    ws = scn.to_mantid(data, param_dim)

    assert len(ws.readX(0)) == expected_bins
    assert ws.getNumberHistograms() == expected_number_spectra
    # check that no workspaces have been leaked in the ADS
    assert len(mtd) == 0, f"Workspaces present: {mtd.getObjectNames()}"

    for i in range(expected_number_spectra):
        np.testing.assert_array_equal(ws.readX(i), x['spectrum', i].values)
        np.testing.assert_array_equal(ws.readY(i), y['spectrum', i].values)
        np.testing.assert_array_equal(ws.readE(i), np.sqrt(y['spectrum', i].variances))


def test_to_workspace_2d_handles_single_spectra():
    from mantid.simpleapi import mtd

    mtd.clear()

    expected_x = [0.0, 1.0, 2.0]
    expected_y = [10.0, 20.0, 30.0]
    expected_e = [4.0, 4.0, 4.0]

    x = sc.Variable(dims=['tof'], values=expected_x)
    y = sc.Variable(dims=['tof'], values=expected_y, variances=expected_e)
    data = sc.DataArray(data=y, coords={'tof': x})

    ws = scn.to_mantid(data, "tof")

    assert ws.getNumberHistograms() == 1

    assert np.equal(ws.readX(0), expected_x).all()
    assert np.equal(ws.readY(0), expected_y).all()
    assert np.equal(ws.readE(0), np.sqrt(expected_e)).all()


def test_to_workspace_2d_handles_single_x_array():
    from mantid.simpleapi import mtd

    mtd.clear()

    expected_x = [0.0, 1.0, 2.0]
    expected_y = [[10.0, 20.0, 30.0], [40.0, 50.0, 60.0]]
    expected_e = [[4.0, 4.0, 4.0], [4.0, 4.0, 4.0]]

    x = sc.Variable(dims=['tof'], values=expected_x)
    y = sc.Variable(
        dims=['spectrum', 'tof'],
        values=np.array(expected_y),
        variances=np.array(expected_e),
    )
    data = sc.DataArray(data=y, coords={'tof': x})

    ws = scn.to_mantid(data, "tof")

    assert ws.getNumberHistograms() == 2
    assert np.equal(ws.readX(0), expected_x).all()
    assert np.equal(ws.readX(1), expected_x).all()

    for i, (y_vals, e_vals) in enumerate(zip(expected_y, expected_e)):
        assert np.equal(ws.readY(i), y_vals).all()
        assert np.equal(ws.readE(i), np.sqrt(e_vals)).all()


def test_attrs_with_dims():
    import mantid.simpleapi as sapi
    from mantid.kernel import FloatArrayProperty

    dataX = [1, 2, 3]
    dataY = [1, 2, 3]
    ws = sapi.CreateWorkspace(DataX=dataX, DataY=dataY, NSpec=1, UnitX="Wavelength")
    # Time series property
    sapi.AddSampleLog(ws, 'attr0', LogText='1', LogType='Number Series')
    # Single value property
    sapi.AddSampleLog(ws, 'attr1', LogText='1', LogType='Number')
    # Array property (not time series)
    p = FloatArrayProperty('attr2', np.arange(10))
    run = ws.mutableRun()
    run.addProperty('attr2', p, replace=True)

    dg = scn.from_mantid(ws)
    assert dg['attr0'].dtype == sc.DType.int32
    assert 'time' in dg['attr0'].coords
    assert dg['attr1'].dtype == sc.DType.int64
    assert dg['attr2'].dtype == sc.DType.float64
    assert dg['attr2'].shape == (10,)


def test_time_series_log_extraction():
    ws = mantid.CreateWorkspace(DataX=[0, 1], DataY=[1])
    times = [
        np.datetime64(t)
        for t in ['2021-01-01T00:00:00', '2021-01-01T00:30:00', '2021-01-01T00:50:00']
    ]
    for i, t in enumerate(times):
        mantid.AddTimeSeriesLog(ws, Name='time_log', Time=str(t), Value=float(i))
    dg = scn.from_mantid(ws)
    assert dg['time_log'].coords['time'].dtype == sc.DType.datetime64
    # check times
    assert sc.identical(
        sc.Variable(dims=['time'], values=np.array(times).astype('datetime64[ns]')),
        dg['time_log'].coords['time'],
    )
    # check values
    assert sc.identical(
        sc.Variable(dims=['time'], values=np.arange(3.0)),
        dg['time_log'].data,
    )
    mantid.DeleteWorkspace(ws)


def test_from_mask_workspace():
    from os import path

    from mantid.simpleapi import LoadMask

    dir_path = path.dirname(path.realpath(__file__))
    mask = LoadMask('HYS', path.join(dir_path, 'HYS_mask.xml'))
    da = scn.from_mantid(mask)['data']
    assert da.unit is None
    assert da.data.dtype == sc.DType.bool
    assert da.dims == ('spectrum',)
    assert da.variances is None


def _all_indirect(blacklist):
    from mantid.simpleapi import config

    # Any indirect instrument considered
    for f in config.getFacilities():
        for i in f.instruments():
            if i.name() not in blacklist and [
                t for t in i.techniques() if 'Indirect' in t
            ]:
                yield i.name()


def _load_indirect_instrument(instr, parameters):
    from mantid.simpleapi import (
        AddSampleLog,
        LoadEmptyInstrument,
        LoadParameterFile,
        config,
    )

    # Create a workspace from an indirect instrument
    out = LoadEmptyInstrument(InstrumentName=instr)
    if instr in parameters:
        LoadParameterFile(
            out,
            Filename=os.path.join(config.getInstrumentDirectory(), parameters[instr]),
        )
    if not out.run().hasProperty('EMode'):
        # EMode would usually get attached via data loading
        # We skip that so have to apply manually
        AddSampleLog(out, LogName='EMode', LogText='Indirect', LogType='String')
    return out


def test_extract_energy_final():
    # Efinal is often stored in a non-default parameter file
    parameters = {
        'IN16B': 'IN16B_silicon_311_Parameters.xml',
        'IRIS': 'IRIS_mica_002_Parameters.xml',
        'OSIRIS': 'OSIRIS_graphite_002_Parameters.xml',
        'BASIS': 'BASIS_silicon_311_Parameters.xml',
    }
    unsupported = ['ZEEMANS', 'MARS', 'IN10', 'IN13', 'IN16', 'VISION', 'VESUVIO']
    for instr in _all_indirect(blacklist=unsupported):
        out = _load_indirect_instrument(instr, parameters)
        da = scn.from_mantid(out)['data']
        efs = da.coords["final_energy"]
        assert not sc.all(sc.isnan(efs)).value
        assert efs.unit == sc.Unit("meV")


def test_extract_energy_final_when_not_present():
    from mantid.kernel import DeltaEModeType
    from mantid.simpleapi import CreateSampleWorkspace

    ws = CreateSampleWorkspace(StoreInADS=False)
    assert ws.getEMode() == DeltaEModeType.Elastic
    da = scn.from_mantid(ws)['data']
    assert "final_energy" not in da.coords


def test_extract_energy_initial():
    from mantid.simpleapi import mtd

    mtd.clear()
    dg = scn.load_with_mantid(
        scn.data.get_path("CNCS_51936_event.nxs"), mantid_args={"SpectrumMax": 1}
    )
    assert sc.identical(
        dg['data'].coords["incident_energy"], sc.scalar(value=3.0, unit=sc.Unit("meV"))
    )


def test_extract_energy_inital_when_not_present():
    from mantid.kernel import DeltaEModeType
    from mantid.simpleapi import CreateSampleWorkspace

    ws = CreateSampleWorkspace(StoreInADS=False)
    assert ws.getEMode() == DeltaEModeType.Elastic
    dg = scn.from_mantid(ws)
    assert "incident_energy" not in dg['data'].coords


def test_EventWorkspace_with_pulse_times():
    small_event_ws = mantid.CreateSampleWorkspace(
        WorkspaceType='Event', NumBanks=1, NumEvents=10
    )
    d = scn.mantid.convert_EventWorkspace_to_data_group(
        small_event_ws, load_pulse_times=True
    )
    assert d['data'].values[0].coords['pulse_time'].dtype == sc.DType.datetime64
    assert sc.identical(
        d['data'].values[0].coords['pulse_time']['event', 0],
        sc.scalar(
            value=small_event_ws.getSpectrum(0).getPulseTimes()[0].to_datetime64()
        ),
    )


@pytest.mark.filterwarnings("ignore:convert_EventWorkspace_to_data_array")
def test_EventWorkspace_with_pulse_times_array():
    small_event_ws = mantid.CreateSampleWorkspace(
        WorkspaceType='Event', NumBanks=1, NumEvents=10
    )
    d = scn.mantid.convert_EventWorkspace_to_data_array(
        small_event_ws, load_pulse_times=True
    )
    assert d.data.values[0].coords['pulse_time'].dtype == sc.DType.datetime64
    assert sc.identical(
        d.data.values[0].coords['pulse_time']['event', 0],
        sc.scalar(
            value=small_event_ws.getSpectrum(0).getPulseTimes()[0].to_datetime64()
        ),
    )


def test_duplicate_monitor_names():
    from mantid.simpleapi import LoadEmptyInstrument

    ws = LoadEmptyInstrument(
        InstrumentName='POLARIS', StoreInADS=False
    )  # Has many monitors named 'monitor'
    monitors = scn.mantid.from_mantid(ws)["monitors"]
    assert monitors['monitor_1']['data'].coords['spectrum'].value == 1
    assert monitors['monitor_13']['data'].coords['spectrum'].value == 13
    assert monitors['monitor_14']['data'].coords['spectrum'].value == 14


def test_load_error_when_file_not_found_via_fuzzy_match():
    with pytest.raises(ValueError):
        scn.load_with_mantid("fictional.nxs")


def make_dynamic_algorithm_without_fileproperty(alg_name):
    from mantid.api import (
        AlgorithmFactory,
        PythonAlgorithm,
        WorkspaceFactory,
        WorkspaceProperty,
    )
    from mantid.kernel import Direction

    # Loader without FileProperty
    if AlgorithmFactory.exists(alg_name):
        return

    class Alg(PythonAlgorithm):
        def PyInit(self):
            self.declareProperty("Filename", "")
            self.declareProperty(
                WorkspaceProperty(
                    name="OutputWorkspace", defaultValue="", direction=Direction.Output
                )
            )

        def PyExec(self):
            self.setProperty("OutputWorkspace", WorkspaceFactory.createTable())

    Alg.__name__ = alg_name
    AlgorithmFactory.subscribe(Alg)
    importlib.reload(sys.modules["mantid.simpleapi"])


def test_load_error_when_file_not_found_via_exact_match():
    make_dynamic_algorithm_without_fileproperty("DummyLoader")
    with pytest.raises(ValueError):
        # DummyLoader has no FileProperty and forces
        # load to evaluate the path given as an absolute path
        scn.load_with_mantid("fictional.nxs", mantid_alg="DummyLoader")


def test_load_via_exact_match():
    make_dynamic_algorithm_without_fileproperty("DummyLoader")
    # scn.load_with_mantid will need to check file exists
    # DummyLoader simply returns a TableWorkspace
    with tempfile.NamedTemporaryFile() as fp:
        scn.load_with_mantid(fp.name, mantid_alg="DummyLoader")
        # Sanity check corrupt full path will fail
        with pytest.raises(ValueError):
            scn.load_with_mantid("fictional_" + fp.name, mantid_alg="DummyLoader")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""
Functions for computing coordinates related to beamline geometry.

Most functions in this module assume a straight beamline geometry.
That is, the beams are not curved by, e.g., a beam guide.
All quantities are defined as in the image below:

.. image:: ../../../docs/_static/straight-beamline.svg
   :scale: 75 %
   :alt: Scattering angle in Bragg reflection.
   :align: center

The 'source position' defines the point where :math:`t=0` (or vice versa).
Neutrons are scattered at the 'sample position' and detected
at 'position' at :math:`t=\\mathsf{tof}`.

In addition, ``total_straight_beam_length_no_scatter`` can be used if the beam
does not scatter off of a sample.
This is useful for instance for monitors.
In this case, ``Ltotal`` is the distance from source to detector.
"""

import scipp as sc
from scipp.typing import VariableLike


def L1(*, incident_beam: VariableLike) -> VariableLike:
    """Compute the length of the incident beam.

    The result is the primary beam length

    .. math::

        L_1 = | \\mathtt{incident\\_beam} |

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`L_1`

    See Also
    --------
    scippneutron.conversions.beamline.straight_incident_beam:
    """
    return sc.norm(incident_beam)


def L2(*, scattered_beam: VariableLike) -> VariableLike:
    """Compute the length of the scattered beam.

    The result is the secondary beam length

    .. math::

        L_2 = | \\mathtt{incident\\_beam} |

    Parameters
    ----------
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`L_2`

    See Also
    --------
    scippneutron.conversions.beamline.straight_scattered_beam:
    """
    return sc.norm(scattered_beam)


def straight_incident_beam(
    *, source_position: VariableLike, sample_position: VariableLike
) -> VariableLike:
    """Compute the length of the beam from source to sample.

    Assumes a straight beam.
    The result is

    .. math::

        \\mathtt{incident\\_beam} = \\mathtt{sample\\_position}
                                    - \\mathtt{source\\_position}

    Parameters
    ----------
    source_position:
        Position of the beam's source.
    sample_position:
        Position of the sample.

    Returns
    -------
    :
        ``incident_beam``
    """
    return sample_position - source_position


def straight_scattered_beam(
    *, position: VariableLike, sample_position: VariableLike
) -> VariableLike:
    """Compute the length of the beam from sample to detector.

    Assumes a straight beam.
    The result is

    .. math::

        \\mathtt{scattered\\_beam} = \\mathtt{position} - \\mathtt{sample\\_position}

    Parameters
    ----------
    position:
        Position of the detector.
    sample_position:
        Position of the sample.

    Returns
    -------
    :
        ``scattered_beam``
    """
    return position - sample_position


def total_beam_length(*, L1: VariableLike, L2: VariableLike) -> VariableLike:
    """Compute the combined length of the incident and scattered beams.

    The result is

    .. math::

        L_\\mathsf{total} = | L_1 + L_2 |

    Parameters
    ----------
    L1:
        Primary path length (incident beam).
    L2:
        Secondary path length (scattered beam).

    Returns
    -------
    :
        :math:`L_\\mathsf{total}`
    """
    return L1 + L2


def total_straight_beam_length_no_scatter(
    *, source_position: VariableLike, position: VariableLike
) -> VariableLike:
    """Compute the length of the beam from source to given position.

    Assumes a straight beam.
    The result is

    .. math::

        L_\\mathsf{total} = | \\mathtt{position} - \\mathtt{{source\\_position}} |

    Parameters
    ----------
    source_position:
        Position of the beam's source. Expects ``dtype=vector3``.
    position:
        Position of the detector. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`L_\\mathsf{total}`
    """
    return sc.norm(position - source_position)


def two_theta(
    *, incident_beam: VariableLike, scattered_beam: VariableLike
) -> VariableLike:
    """Compute the scattering angle between scattered and transmitted beams.

    See :mod:`scippneutron.conversions.beamline` for the definition of the angle.

    The result is equivalent to

    .. math::

        b_1 &= \\mathtt{incident\\_beam} / |\\mathtt{incident\\_beam}| \\\\
        b_2 &= \\mathtt{scattered\\_beam} / |\\mathtt{scattered\\_beam}| \\\\
        2\\theta &= \\mathsf{acos}(b_1 \\cdot b_2)

    but uses a numerically more stable implementation by W. Kahan
    (https://people.eecs.berkeley.edu/~wkahan/MathH110/Cross.pdf).

    Parameters
    ----------
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.

    Returns
    -------
    :
        :math:`2\\theta`

    See Also
    --------
    scippneutron.conversions.beamline.straight_incident_beam:
    scippneutron.conversions.beamline.straight_scattered_beam:
    """
    # The implementation is based on paragraph 13 of
    # https://people.eecs.berkeley.edu/~wkahan/MathH110/Cross.pdf
    # Which is Kahan:2000:CPR in
    # https://netlib.org/bibnet/authors/k/kahan-william-m.html
    # And referenced by https://scicomp.stackexchange.com/a/27769
    #
    # It claims that the formula is 'valid for euclidean spaces of any dimension'
    # and 'it never errs by more than a modest multiple of epsilon'
    # where 'epsilon is the roundoff threshold for individual arithmetic
    # operations'.
    b1 = incident_beam / L1(incident_beam=incident_beam)
    b2 = scattered_beam / L2(scattered_beam=scattered_beam)
    return 2 * sc.atan2(y=sc.norm(b1 - b2), x=sc.norm(b1 + b2))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pathlib
import sys
import warnings
from pathlib import Path
from typing import Callable, List, Type, Union

import numpy as np
import pytest
import scipp as sc
from dateutil.parser import parse as parse_date

import scippneutron
from scippneutron._utils import get_attrs, get_meta
from scippneutron.io.nexus.load_nexus import load_nexus_json_str

sys.path.insert(0, str(Path(__file__).resolve().parent))
from nexus_helpers import (  # noqa: E402
    Chopper,
    Detector,
    EventData,
    Link,
    Log,
    Monitor,
    NexusBuilder,
    Sample,
    Source,
    Transformation,
    TransformationType,
    in_memory_hdf5_file_with_two_nxentry,
)

# representative sample of UTF-8 test strings from
# https://www.w3.org/2001/06/utf-8-test/UTF-8-demo.html
UTF8_TEST_STRINGS = (
    "∮ E⋅da = Q,  n → ∞, ∑ f(i) = ∏ g(i), ∀x∈ℝ: ⌈x⌉ = −⌊−x⌋, α ∧ ¬β = ¬(¬α ∨ β)",
    "2H₂ + O₂ ⇌ 2H₂O, R = 4.7 kΩ, ⌀ 200 mm",
    "Σὲ γνωρίζω ἀπὸ τὴν κόψη",
)


def _timestamp(date: str):
    return parse_date(date).timestamp()


def test_raises_exception_if_multiple_nxentry_in_file():
    with in_memory_hdf5_file_with_two_nxentry() as nexus_file:
        with pytest.raises(RuntimeError):
            scippneutron.load_nexus(nexus_file)


def test_no_exception_if_single_nxentry_found_below_root():
    with in_memory_hdf5_file_with_two_nxentry() as nexus_file:
        # There are 2 NXentry in the file, but root is used
        # to specify which to load data from
        assert scippneutron.load_nexus(nexus_file, root='/entry_1') is None


def load_from_nexus(
    builder: NexusBuilder, *args, **kwargs
) -> Union[sc.Dataset, sc.DataArray, None]:
    with builder.file() as nexus_file:
        return scippneutron.load_nexus(nexus_file, *args, **kwargs)


def load_from_json(
    builder: NexusBuilder, *args, **kwargs
) -> Union[sc.Dataset, sc.DataArray, None]:
    loaded_data, _ = load_nexus_json_str(builder.json_string, *args, **kwargs)
    return loaded_data


@pytest.fixture(params=[load_from_nexus, load_from_json])
def load_function(request) -> Callable:
    """
    Each test with this fixture is executed with load_nexus_json
    loading JSON output from the NexusBuilder, and with load_nexus
    loading in-memory NeXus output from the NexusBuilder
    """
    return request.param


@pytest.mark.skip(
    reason="The 'bigfake' file is partially broken and contains "
    "HDF5 groups without NX_class attribute."
)
@pytest.mark.parametrize('path_type', (str, pathlib.Path))
def test_loads_from_file(path_type):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message='Skipped load', category=UserWarning)
        da = scippneutron.load_nexus(path_type(scippneutron.data.bigfake()))
    assert da.sizes == {'detector_id': 90000, 'tof': 1}
    assert 'tof' in da.coords
    assert 'position' in da.coords


def test_no_exception_if_single_nxentry_in_file(load_function: Callable):
    builder = NexusBuilder()
    assert load_function(builder) is None


def test_loads_data_from_single_event_data_group(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)

    loaded_data = load_function(builder)

    assert loaded_data.bins.constituents['data'].unit == 'counts'

    # Expect time of flight to match the values in the
    # event_time_offset dataset
    # May be reordered due to binning (hence np.sort)
    assert sc.identical(
        sc.sort(
            loaded_data.bins.concat('detector_id').values[0].coords['tof'], key="event"
        ),
        sc.sort(
            sc.array(dims=["event"], values=event_time_offsets, unit=sc.units.ns),
            key="event",
        ),
    )

    counts_on_detectors = loaded_data.bins.sum()
    # No detector_number dataset in file so expect detector_id to be
    # binned according to whatever detector_ids are present in event_id
    # dataset: 2 on det 1, 1 on det 2, 2 on det 3
    expected_counts = np.array([[2], [1], [2]])
    assert sc.identical(
        counts_on_detectors.data,
        sc.array(
            dims=['detector_id', 'tof'],
            unit='counts',
            dtype='float32',
            values=expected_counts,
        ),
    )
    expected_detector_ids = np.array([1, 2, 3])
    assert np.array_equal(
        loaded_data.coords['detector_id'].values, expected_detector_ids
    )


@pytest.mark.parametrize(
    "unit,multiplier", (("ns", 1), ("us", 10**3), ("ms", 10**6), ("s", 10**9))
)
def test_loads_pulse_times_from_single_event_with_different_units(
    load_function: Callable, unit: str, multiplier: float
):
    offsets = np.array([12, 34, 56, 78])
    zeros = np.array([12.0, 34.0, 56.0, 78.0], dtype="float64")
    event_data = EventData(
        event_id=np.array([1, 2, 3, 4]),
        event_time_offset=offsets,
        event_time_zero=zeros,
        event_index=np.array([0, 3, 3, 4]),
        event_time_zero_unit=unit,
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([1, 2, 3, 4]), event_data=event_data)
    )

    loaded_data = load_function(builder)

    for event, pulse_time in enumerate([12, 12, 12, 56]):
        _time = (
            np.array("1970-01-01").astype("datetime64[ns]")
            + np.array(pulse_time).astype("timedelta64[ns]") * multiplier
        )

        # Allow 1ns difference for rounding errors between different routes
        assert all(
            np.abs(loaded_data.values[event].coords['pulse_time'].values - _time)
            <= np.array(1).astype("timedelta64[ns]")
        )


@pytest.mark.parametrize(
    "time_zero_offset,time_zero,time_zero_unit,expected_time",
    (
        ("1980-01-01T00:00:00.0", 30, "s", "1980-01-01T00:00:30.0"),
        ("1990-01-01T00:00:00.0", 5000, "ms", "1990-01-01T00:00:05.0"),
        ("2000-01-01T00:00:00.0", 3 * 10**6, "us", "2000-01-01T00:00:03.0"),
        ("2010-01-01T00:00:00.0", 12, "hour", "2010-01-01T12:00:00.0"),
    ),
)
def test_loads_pulse_times_with_combinations_of_offset_and_units(
    load_function: Callable,
    time_zero_offset: str,
    time_zero: float,
    time_zero_unit: str,
    expected_time: str,
):
    offsets = np.array([0])
    zeros = np.array([time_zero], dtype="float64")
    event_data = EventData(
        event_id=np.array([0]),
        event_time_offset=offsets,
        event_time_zero_offset=time_zero_offset,
        event_time_zero=zeros,
        event_index=np.array([0]),
        event_time_zero_unit=time_zero_unit,
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([0]), event_data=event_data)
    )

    loaded_data = load_function(builder)

    _time = np.array(expected_time).astype("datetime64[ns]")

    # Allow 1ns difference for rounding errors between different routes
    assert np.abs(
        loaded_data.values[0].coords['pulse_time'].values[0] - _time
    ) <= np.array(1).astype("timedelta64[ns]")


def test_does_not_load_events_if_index_not_ordered(load_function: Callable):
    event_data_1 = EventData(
        event_id=np.array([0, 1]),
        event_time_offset=np.array([0, 1]),
        event_time_zero=np.array([0, 1]),
        event_index=np.array([2, 0]),
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([0, 1]), event_data=event_data_1)
    )

    with pytest.warns(UserWarning, match="Invalid index in NXevent_data at "):
        load_function(builder)


def test_loads_pulse_times_from_multiple_event_data_groups(load_function: Callable):
    offsets = np.array([0, 0, 0, 0])

    zeros_1 = np.array([12 * 10**9, 34 * 10**9, 56 * 10**9, 78 * 10**9])
    zeros_2 = np.array([87 * 10**9, 65 * 10**9, 43 * 10**9, 21 * 10**9])

    event_data_1 = EventData(
        event_id=np.array([0, 1, 2, 3]),
        event_time_offset=offsets,
        event_time_zero=zeros_1,
        event_index=np.array([0, 3, 3, 4]),
    )
    event_data_2 = EventData(
        event_id=np.array([4, 5, 6, 7]),
        event_time_offset=offsets,
        event_time_zero=zeros_2,
        event_index=np.array([0, 3, 3, 4]),
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([0, 1, 2, 3]), event_data=event_data_1)
    )
    builder.add_detector(
        Detector(detector_numbers=np.array([4, 5, 6, 7]), event_data=event_data_2)
    )

    loaded_data = load_function(builder)

    for event, pulse_time in enumerate([12, 12, 12, 56, 87, 87, 87, 43]):
        _time = np.array("1970-01-01").astype("datetime64[ns]") + np.array(
            pulse_time
        ).astype("timedelta64[s]")

        assert sc.identical(
            loaded_data.values[event].coords['pulse_time'],
            sc.array(
                dims=["event"],
                values=[_time],
                unit=sc.units.ns,
                dtype=sc.DType.datetime64,
            ),
        )


def test_loads_data_from_multiple_event_data_groups(load_function: Callable):
    pulse_times = np.array(
        [
            1600766730000000000,
            1600766731000000000,
            1600766732000000000,
            1600766733000000000,
        ]
    )
    event_time_offsets_1 = np.array([456, 743, 347, 345, 632])
    event_data_1 = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets_1,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_1_ids = np.array([0, 1, 2, 3])
    event_time_offsets_2 = np.array([682, 237, 941, 162, 52])
    event_data_2 = EventData(
        event_id=np.array([4, 5, 6, 4, 6]),
        event_time_offset=event_time_offsets_2,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_2_ids = np.array([4, 5, 6, 7])

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_1_ids, event_data_1))
    builder.add_detector(Detector(detector_2_ids, event_data_2))

    loaded_data = load_function(builder)

    # Expect time of flight to match the values in the
    # event_time_offset datasets
    # May be reordered due to binning (hence np.sort)
    assert np.array_equal(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(np.concatenate((event_time_offsets_1, event_time_offsets_2))),
    )

    counts_on_detectors = loaded_data.bins.sum()
    # There are detector_number datasets in the NXdetector for each
    # NXevent_data, these are used for detector_id binning
    expected_counts = np.array([[0], [2], [1], [2], [2], [1], [2], [0]])
    assert np.array_equal(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.concatenate((detector_1_ids, detector_2_ids))
    assert np.array_equal(
        loaded_data.coords['detector_id'].values, expected_detector_ids
    )


def test_skips_event_data_group_with_non_integer_event_ids(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1.1, 2.2, 3.3, 1.1, 3.1]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None, (
        "Expected no data to be loaded as " "event data has non integer event ids"
    )


def test_skips_event_data_group_with_non_integer_detector_numbers(
    load_function: Callable,
):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_numbers = np.array([0.1, 1.2, 2.3, 3.4])

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_numbers, event_data))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None, (
        "Expected no data to be loaded as " "detector has non integer detector numbers"
    )


def test_skips_data_with_event_id_and_detector_number_type_unequal(
    load_function: Callable,
):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3], dtype=np.int64),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_numbers = np.array([0, 1, 2, 3], dtype=np.float32)

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_numbers, event_data))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None, (
        "Expected no data to be loaded as event "
        "ids and detector numbers are of "
        "different types"
    )


def test_loads_data_from_single_log_with_no_units(load_function: Callable):
    values = np.array([1, 2, 3])
    times = np.array([4, 5, 6])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values, times))

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert np.array_equal(loaded_data[name].data.values.values, values)


def test_loads_data_from_single_log_with_units(load_function: Callable):
    values = np.array([1.1, 2.2, 3.3])
    times = np.array([4.4, 5.5, 6.6])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values, times, value_units="m", time_units="s"))

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert np.allclose(loaded_data[name].data.values.values, values)
    assert loaded_data[name].data.values.unit == sc.units.m


def test_loads_data_from_multiple_logs(load_function: Callable):
    builder = NexusBuilder()
    log_1 = Log("test_log", np.array([1.1, 2.2, 3.3]), np.array([4.4, 5.5, 6.6]))
    log_2 = Log("test_log_2", np.array([123, 253, 756]), np.array([246, 1235, 2369]))
    builder.add_log(log_1)
    builder.add_log(log_2)

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert np.allclose(loaded_data[log_1.name].data.values.values, log_1.value)
    assert np.array_equal(loaded_data[log_2.name].data.values.values, log_2.value)


def test_loads_logs_with_non_supported_int_types(load_function: Callable):
    builder = NexusBuilder()
    log_int8 = Log(
        "test_log_int8", np.array([1, 2, 3]).astype(np.int8), np.array([4.4, 5.5, 6.6])
    )
    log_int16 = Log(
        "test_log_int16",
        np.array([123, 253, 756]).astype(np.int16),
        np.array([246, 1235, 2369]),
    )
    log_uint8 = Log(
        "test_log_uint8",
        np.array([1, 2, 3]).astype(np.uint8),
        np.array([4.4, 5.5, 6.6]),
    )
    log_uint16 = Log(
        "test_log_uint16",
        np.array([123, 253, 756]).astype(np.uint16),
        np.array([246, 1235, 2369]),
    )
    logs = (log_int8, log_int16, log_uint8, log_uint16)
    for log in logs:
        builder.add_log(log)

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    for log in logs:
        assert np.allclose(loaded_data[log.name].data.values.values, log.value)


def test_loads_multidimensional_log(load_function: Callable):
    multidim_values = np.array([[1, 2, 3], [1, 2, 3]])
    times = np.array([4, 5])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, multidim_values, times, value_units=''))

    loaded_data = load_function(builder)

    log = loaded_data[name].data.value
    expected = sc.DataArray(sc.array(dims=['time', 'dim_1'], values=multidim_values))
    expected.coords['time'] = sc.epoch(unit='s') + sc.array(
        dims=['time'], values=times, unit='s'
    )
    assert sc.identical(log, expected)


def test_skips_log_with_no_value_dataset(load_function: Callable):
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, None, np.array([4, 5, 6])))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert isinstance(loaded_data['test_log'].value, sc.DataGroup)


def test_loads_log_with_empty_value_and_time_datasets(load_function: Callable):
    empty_values = np.array([]).astype(np.int32)
    empty_times = np.array([]).astype(np.int32)
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, empty_values, empty_times))

    loaded_data = load_function(builder)

    assert loaded_data[name].data.values.sizes == {'time': 0}


def test_warns_given_log_with_mismatched_value_and_time(load_function: Callable):
    values = np.array([1, 2, 3]).astype(np.int32)
    # Note that if times exceeds length by 1 it is loaded as bin edges. It is unclear
    # if this is considered valid Nexus.
    times = np.array([1, 2, 3, 4, 5]).astype(np.int32)
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values, times))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert set(loaded_data['test_log'].value.keys()) == {'time', 'value'}


def test_loads_data_from_non_timeseries_log(load_function: Callable):
    values = np.array([1.1, 2.2, 3.3])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values))

    loaded_data = load_function(builder)

    assert np.allclose(loaded_data[name].data.values.values, values)


def test_loads_data_from_multiple_logs_with_same_name(load_function: Callable):
    values_1 = np.array([1.1, 2.2, 3.3])
    values_2 = np.array([4, 5, 6])
    name = "test_log"

    # Add one log to NXentry and the other to an NXdetector,
    # both have the same group name
    builder = NexusBuilder()
    builder.add_log(Log(name, values_1))
    builder.add_detector(Detector(log=Log(name, values_2)))

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message='Failed to load', category=UserWarning
        )
        warnings.filterwarnings(
            "ignore", message='Skipped loading', category=UserWarning
        )
        loaded_data = load_function(builder)

    assert len(loaded_data) == 2
    for key, da in loaded_data.items():
        assert key.endswith('test_log')
        if 'detector_0' in key:
            assert np.allclose(da.value.values, values_2)
        else:
            assert np.allclose(da.value.values, values_1)


def test_load_instrument_name(load_function: Callable):
    name = "INSTR"
    builder = NexusBuilder()
    builder.add_instrument(name)

    loaded_data = load_function(builder)

    assert loaded_data['instrument_name'].values == name


def test_load_experiment_title(load_function: Callable):
    title = "my experiment"
    builder = NexusBuilder()
    builder.add_title(title)

    loaded_data = load_function(builder)

    assert loaded_data['experiment_title'].values == title


def test_loads_event_and_log_data_from_single_file(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    log_1 = Log("test_log", np.array([1.1, 2.2, 3.3]), np.array([4.4, 5.5, 6.6]))
    log_2 = Log("test_log_2", np.array([123, 253, 756]), np.array([246, 1235, 2369]))

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    builder.add_log(log_1)
    builder.add_log(log_2)

    loaded_data = load_function(builder)

    # Expect time of flight to match the values in the
    # event_time_offset dataset
    # May be reordered due to binning (hence np.sort)
    assert np.allclose(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(event_time_offsets),
    )

    counts_on_detectors = loaded_data.bins.sum()
    # No detector_number dataset in file so expect detector_id to be
    # binned from the min to the max detector_id recorded in event_id
    # dataset: 2 on det 1, 1 on det 2, 2 on det 3
    expected_counts = np.array([[2], [1], [2]])
    assert np.allclose(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.array([1, 2, 3])
    assert np.allclose(loaded_data.coords['detector_id'].values, expected_detector_ids)
    assert "position" not in get_meta(loaded_data).keys(), (
        "The NXdetectors had no pixel position datasets so we "
        "should not find 'position' coord"
    )

    # Logs should have been added to the DataArray as attributes
    assert np.allclose(get_attrs(loaded_data)[log_1.name].values.values, log_1.value)
    assert np.allclose(get_attrs(loaded_data)[log_2.name].values.values, log_2.value)


def test_loads_pixel_positions_with_event_data(load_function: Callable):
    pulse_times = np.array(
        [
            1600766730000000000,
            1600766731000000000,
            1600766732000000000,
            1600766733000000000,
        ]
    )
    event_time_offsets_1 = np.array([456, 743, 347, 345, 632])
    event_data_1 = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets_1,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_1_ids = np.array([0, 1, 2, 3])
    x_pixel_offset_1 = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset_1 = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset_1 = np.array([0.1, 0.2, 0.3, 0.4])

    event_time_offsets_2 = np.array([682, 237, 941, 162, 52])
    event_data_2 = EventData(
        event_id=np.array([4, 5, 6, 4, 6]),
        event_time_offset=event_time_offsets_2,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    # Multidimensional is fine as long as the shape of
    # the ids and the pixel offsets match
    detector_2_ids = np.array([[4, 5], [6, 7]])
    x_pixel_offset_2 = np.array([[1.1, 1.2], [1.1, 1.2]])
    y_pixel_offset_2 = np.array([[0.1, 0.1], [0.2, 0.2]])

    builder = NexusBuilder()
    offsets_units = "mm"
    builder.add_detector(
        Detector(
            detector_1_ids,
            event_data_1,
            x_offsets=x_pixel_offset_1,
            y_offsets=y_pixel_offset_1,
            z_offsets=z_pixel_offset_1,
            offsets_unit=offsets_units,
        )
    )
    builder.add_detector(
        Detector(
            detector_2_ids,
            event_data_2,
            x_offsets=x_pixel_offset_2,
            y_offsets=y_pixel_offset_2,
            offsets_unit=offsets_units,
        )
    )

    loaded_data = load_function(builder)

    # If z offsets are missing they should be zero
    z_pixel_offset_2 = np.array([[0.0, 0.0], [0.0, 0.0]])
    expected_pixel_positions = np.array(
        [
            np.concatenate((x_pixel_offset_1, x_pixel_offset_2.flatten())),
            np.concatenate((y_pixel_offset_1, y_pixel_offset_2.flatten())),
            np.concatenate((z_pixel_offset_1, z_pixel_offset_2.flatten())),
        ]
    ).T
    assert np.allclose(loaded_data.coords['position'].values, expected_pixel_positions)
    assert get_meta(loaded_data)['position'].unit == 'mm'


@pytest.mark.skip(
    reason="It is unclear if NXdetector with neither data nor event_data "
    "needs to be considered valid."
)
def test_loads_pixel_positions_without_event_data(load_function: Callable):
    """
    This is important in the live-data feature as geometry and event data
    are streamed separately
    """
    detector_1_ids = np.array([0, 1, 2, 3])
    x_pixel_offset_1 = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset_1 = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset_1 = np.array([0.1, 0.2, 0.3, 0.4])

    detector_2_ids = np.array([[4, 5], [6, 7]])
    x_pixel_offset_2 = np.array([[1.1, 1.2], [1.1, 1.2]])
    y_pixel_offset_2 = np.array([[0.1, 0.1], [0.2, 0.2]])

    builder = NexusBuilder()
    offsets_units = "mm"
    builder.add_detector(
        Detector(
            detector_numbers=detector_1_ids,
            x_offsets=x_pixel_offset_1,
            y_offsets=y_pixel_offset_1,
            z_offsets=z_pixel_offset_1,
            offsets_unit=offsets_units,
        )
    )
    builder.add_detector(
        Detector(
            detector_numbers=detector_2_ids,
            x_offsets=x_pixel_offset_2,
            y_offsets=y_pixel_offset_2,
            offsets_unit=offsets_units,
        )
    )

    loaded_data = load_function(builder)

    # If z offsets are missing they should be zero
    z_pixel_offset_2 = np.array([[0.0, 0.0], [0.0, 0.0]])
    expected_pixel_positions = (
        np.array(
            [
                np.concatenate((x_pixel_offset_1, x_pixel_offset_2.flatten())),
                np.concatenate((y_pixel_offset_1, y_pixel_offset_2.flatten())),
                np.concatenate((z_pixel_offset_1, z_pixel_offset_2.flatten())),
            ]
        ).T
        / 1_000
    )  # Divide by 1000 for mm to metres
    assert np.allclose(loaded_data.coords['position'].values, expected_pixel_positions)
    assert (
        get_meta(loaded_data)['base_position'].unit == sc.units.m
    ), "Expected positions to be converted to metres"


def test_loads_pixel_positions_with_no_units(load_function: Callable):
    pulse_times = np.array(
        [
            1600766730000000000,
            1600766731000000000,
            1600766732000000000,
            1600766733000000000,
        ]
    )
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_ids = np.array([0, 1, 2, 3])
    x_pixel_offset = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset = np.array([0.1, 0.2, 0.3, 0.4])

    builder = NexusBuilder()
    builder.add_detector(
        Detector(
            detector_ids,
            event_data,
            x_offsets=x_pixel_offset,
            y_offsets=y_pixel_offset,
            z_offsets=z_pixel_offset,
            offsets_unit=None,
        )
    )

    loaded_data = load_function(builder)

    assert loaded_data.coords['position'].unit is None


def test_sample_position_at_origin_if_not_explicit_in_file(load_function: Callable):
    # The sample position is the origin of the coordinate
    # system in NeXus files.
    # If there is an NXsample in the file, but it has no "distance" dataset
    # or "depends_on" pointing to NXtransformations then it should be
    # assumed to be at the origin.
    builder = NexusBuilder()
    builder.add_sample(Sample("sample"))
    loaded_data = load_function(builder)

    origin = np.array([0, 0, 0])
    assert np.allclose(loaded_data["sample_position"].values, origin)


def test_loads_multiple_samples(load_function: Callable):
    # More than one sample in the file is possible, but we cannot guess
    # which to record as _the_ "sample_position",
    # instead record them in the form "<GROUP_NAME>_position"
    builder = NexusBuilder()
    sample_1_name = "sample_1"
    sample_2_name = "sample_2"
    builder.add_sample(Sample(sample_1_name))

    distance = 0.762
    units = "m"
    builder.add_sample(Sample(sample_2_name, distance=distance, distance_units=units))
    loaded_data = load_function(builder)

    origin = np.array([0, 0, 0])
    assert np.allclose(loaded_data[f"{sample_1_name}_position"].values, origin), (
        "Sample did not have explicit location so expect position "
        "to be recorded as the origin"
    )
    expected_position = np.array([0, 0, distance])
    assert np.allclose(
        loaded_data[f"{sample_2_name}_position"].values, expected_position
    )


def test_skips_loading_source_if_more_than_one_in_file(load_function: Callable):
    # More than one source is a serious error in the file, so load_nexus will not
    # define a source_position rather than guessing which is the "correct" one.
    builder = NexusBuilder()
    builder.add_source(Source("source_1"))
    builder.add_source(Source("source_2"))
    loaded_data = load_function(builder)
    assert 'source_1' in loaded_data
    assert 'source_2' in loaded_data
    assert 'source_position' not in loaded_data


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
def test_component_position_from_distance_dataset_missing_unit(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    load_function: Callable,
):
    builder = NexusBuilder()
    distance = 4.2
    builder.add_component(
        component_class(component_name, distance=distance, distance_units=None)
    )
    loaded_data = load_function(builder)
    assert loaded_data[f'{component_name}_position'].unit is None


@pytest.mark.parametrize(
    "component_class,component_name", [(Sample, "sample"), (Source, "source")]
)
@pytest.mark.parametrize(
    "transform_type,value,value_units,expected_position",
    (
        (TransformationType.ROTATION, 0.27, "rad", [0, 0, 0]),
        (TransformationType.TRANSLATION, 230, "cm", [0, 0, 2.3]),
    ),
)
def test_loads_component_position_from_single_transformation(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    value: float,
    value_units: str,
    expected_position: List[float],
    load_function: Callable,
):
    builder = NexusBuilder()
    transformation = Transformation(
        transform_type,
        vector=np.array([0, 0, 1]),
        value=np.array(value),
        value_units=value_units,
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)
    assert np.allclose(
        loaded_data[f"{component_name}_position"].values, expected_position
    )
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize(
    "component_class,component_name", [(Sample, "sample"), (Source, "source")]
)
@pytest.mark.parametrize(
    "transform_type,value,value_units,expected_position",
    (
        (TransformationType.ROTATION, 180, "deg", [-1, -2, 3]),
        (TransformationType.TRANSLATION, 230, "cm", [1, 2, 5.3]),
    ),
)
def test_loads_component_position_from_single_transformation_with_offset(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    value: float,
    value_units: str,
    expected_position: List[float],
    load_function: Callable,
):
    builder = NexusBuilder()
    transformation = Transformation(
        transform_type,
        vector=np.array([0, 0, 1]),
        value=np.array(value),
        value_units=value_units,
        offset=[1, 2, 3],
        offset_unit='m',
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)
    assert np.allclose(
        loaded_data[f"{component_name}_position"].values, expected_position
    )
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize(
    "component_class,component_name", [(Sample, "sample"), (Source, "source")]
)
@pytest.mark.parametrize(
    "transform_type,value,value_units,expected_position",
    (
        (TransformationType.ROTATION, 180, "deg", [-1, -2, 3]),
        (TransformationType.TRANSLATION, 230, "cm", [1, 2, 5.3]),
    ),
)
def test_raises_if_offset_but_not_offset_units_found(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    value: float,
    value_units: str,
    expected_position: List[float],
    load_function: Callable,
):
    builder = NexusBuilder()
    transformation = Transformation(
        transform_type,
        vector=np.array([0, 0, 1]),
        value=np.array(value),
        value_units=value_units,
        offset=[1, 2, 3],
        offset_unit=None,
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message='Failed to load', category=UserWarning
        )
        loaded_data = load_function(builder)
    assert isinstance(loaded_data[component_name].value['depends_on'], str)


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
@pytest.mark.parametrize(
    "transform_type,value,value_units,expected_position",
    (
        (TransformationType.ROTATION, 0.27, "rad", [0, 0, 0]),
        (TransformationType.TRANSLATION, 230, "cm", [0, 0, 2.3]),
    ),
)
def test_loads_component_position_from_log_transformation(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    value: float,
    value_units: str,
    expected_position: List[float],
    load_function: Callable,
):
    builder = NexusBuilder()
    # Provide "time" data, the builder will write the transformation as
    # an NXlog
    transformation = Transformation(
        transform_type,
        vector=np.array([0, 0, 1]),
        value=np.array(value),
        value_units=value_units,
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)

    # Should load as usual despite the transformation being an NXlog
    # as it only has a single value
    assert np.allclose(
        loaded_data[f"{component_name}_position"].values, expected_position
    )
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
@pytest.mark.parametrize(
    "transform_type,value,value_units,expected_position",
    (
        (TransformationType.ROTATION, [0.27, 0.73], "rad", [0, 0, 0]),
        (TransformationType.TRANSLATION, [230, 310], "cm", [0, 0, 2.3]),
    ),
)
def test_loads_component_position_with_multi_value_log_transformation(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    value: List[float],
    value_units: str,
    expected_position: float,
    load_function: Callable,
):
    builder = NexusBuilder()
    # Provide "time" data, the builder will write the transformation as
    # an NXlog. This would be encountered in a file from an experiment
    # involving a scan of a motion axis.
    transformation = Transformation(
        transform_type,
        vector=np.array([0, 0, 1]),
        value=np.array(value),
        time=np.array([1.3, 6.4]),
        time_units="s",
        value_units=value_units,
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)
    assert sc.identical(
        loaded_data[component_name].value['depends_on'].coords["time"],
        sc.Variable(
            dims=["time"],
            values=[1300000000, 6400000000],
            unit="ns",
            dtype=sc.DType.datetime64,
        ),
    )


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
def test_loads_component_position_with_multiple_multi_valued_log_transformations(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    load_function: Callable,
):
    builder = NexusBuilder()
    # Provide "time" data, the builder will write the transformation as
    # an NXlog. This would be encountered in a file from an experiment
    # involving a scan of a motion axis.
    t1 = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array([1, 10]),
        time=np.array([0, 1]),
        time_units="s",
        value_units="m",
    )

    t2 = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array([5, 50]),
        time=np.array([0, 1]),
        time_units="s",
        value_units="m",
        depends_on=t1,
    )

    builder.add_component(component_class(component_name, depends_on=t2))
    loaded_data = load_function(builder)
    comp = loaded_data[component_name].value
    assert sc.identical(
        comp['depends_on'].coords["time"],
        sc.Variable(dims=["time"], values=[0, 1], unit="s", dtype=sc.DType.datetime64),
    )


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
def test_multi_valued_log_transformations_time_axis_interpolated_and_trimmed(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    load_function: Callable,
):
    builder = NexusBuilder()
    t1 = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array([1, 10]),
        time=np.array([0, 1]),
        time_units="s",
        value_units="m",
    )
    t2 = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array([5, 50]),
        time=np.array([0.1, 1.1]),
        time_units="s",
        value_units="m",
        depends_on=t1,
    )
    builder.add_component(component_class(component_name, depends_on=t2))

    loaded_data = load_function(builder)

    # Note the start at 100 ms, since there is no known value of t2 at 0 ms (when
    # t1 starts)
    comp = loaded_data[component_name].value
    assert sc.identical(
        comp['depends_on'].coords["time"],
        sc.array(
            dims=["time"],
            values=[100, 1000, 1100],
            unit="ms",
            dtype=sc.DType.datetime64,
        ).to(unit="ns"),
    )


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
@pytest.mark.parametrize(
    "transform_type,value_units",
    ((TransformationType.ROTATION, "deg"), (TransformationType.TRANSLATION, "cm")),
)
def test_skips_component_position_with_empty_value_log_transformation(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    value_units: str,
    load_function: Callable,
):
    builder = NexusBuilder()
    empty_value = np.array([])
    transformation = Transformation(
        transform_type,
        vector=np.array([0, 0, 1]),
        value=empty_value,
        time=np.array([1.3, 6.4]),
        time_units="s",
        value_units=value_units,
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    with pytest.warns(UserWarning):
        load_function(builder)


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
def test_load_component_position_prefers_transform_over_distance(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    load_function: Callable,
):
    # The "distance" dataset gives the position along the z axis.
    # If there is a "depends_on" pointing to transformations then we
    # prefer to use that instead as it is likely to be more accurate; it
    # can define position and orientation in 3D.
    builder = NexusBuilder()
    transformation = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array(2.3),
        value_units="m",
    )
    builder.add_component(
        component_class(
            component_name, depends_on=transformation, distance=4.2, distance_units="m"
        )
    )
    loaded_data = load_function(builder)

    expected_position = np.array([0, 0, transformation.value])
    assert np.allclose(
        loaded_data[f"{component_name}_position"].values, expected_position
    )
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
@pytest.mark.parametrize(
    "transform_type", (TransformationType.ROTATION, TransformationType.TRANSLATION)
)
def test_skips_component_position_from_transformation_missing_unit(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    transform_type: TransformationType,
    load_function: Callable,
):
    builder = NexusBuilder()
    transformation = Transformation(
        transform_type, np.array([0, 0, -1]), value=np.array(2.3)
    )
    builder.add_component(component_class(component_name, depends_on=transformation))
    with pytest.warns(UserWarning):
        load_function(builder)


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
def test_loads_component_position_from_multiple_transformations(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    load_function: Callable,
):
    builder = NexusBuilder()
    transformation_1 = Transformation(
        TransformationType.ROTATION,
        np.array([0, 1, 0]),
        value=np.array(90),
        value_units="deg",
    )
    transformation_2 = Transformation(
        TransformationType.TRANSLATION,
        np.array([0, 0, 1]),
        value=np.array(2.3),
        value_units="m",
        depends_on=transformation_1,
    )
    builder.add_component(component_class(component_name, depends_on=transformation_2))
    loaded_data = load_function(builder)

    expected_position = np.array([2.3, 0, 0])
    assert np.allclose(
        loaded_data[f"{component_name}_position"].values, expected_position
    )
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


def test_skips_source_position_if_not_given_in_file(load_function: Callable):
    builder = NexusBuilder()
    builder.add_source(Source("source"))
    loaded_data = load_function(builder)
    assert 'source_position' not in loaded_data


@pytest.mark.parametrize(
    "component_class,component_name", ((Sample, "sample"), (Source, "source"))
)
def test_loads_component_position_from_distance_dataset(
    component_class: Union[Type[Source], Type[Sample]],
    component_name: str,
    load_function: Callable,
):
    # If the NXsource or NXsample contains a "distance" dataset
    # this gives the position along the z axis. If there was a "depends_on"
    # pointing to transformations then we'd use that instead as it is
    # likely to be more accurate; it can define position and orientation in 3D.
    builder = NexusBuilder()
    distance = 4.2
    units = "m"
    builder.add_component(
        component_class(component_name, distance=distance, distance_units=units)
    )
    loaded_data = load_function(builder)

    expected_position = np.array([0, 0, distance])
    assert np.allclose(
        loaded_data[f"{component_name}_position"].values, expected_position
    )
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit(units)


def test_loads_source_position_dependent_on_sample_position(load_function: Callable):
    builder = NexusBuilder()
    transformation_0 = Transformation(
        TransformationType.ROTATION,
        np.array([0, 1, 0]),
        value=np.array(90),
        value_units="deg",
    )
    transformation_1 = Transformation(
        TransformationType.TRANSLATION,
        np.array([0, 0, 1]),
        value=np.array(2.3),
        value_units="m",
        depends_on=transformation_0,
    )
    builder.add_sample(Sample("sample", depends_on=transformation_1))
    transformation_2 = Transformation(
        TransformationType.TRANSLATION,
        np.array([0, 0, 1]),
        value=np.array(1.0),
        value_units="m",
        depends_on="/entry/sample/transformations/transform_1",
    )
    builder.add_source(Source("source", depends_on=transformation_2))
    loaded_data = load_function(builder)

    expected_position = np.array([3.3, 0, 0])
    assert np.allclose(loaded_data["source_position"].values, expected_position)
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_loads_pixel_positions_with_transformations(load_function: Callable):
    pulse_times = np.array(
        [
            1600766730000000000,
            1600766731000000000,
            1600766732000000000,
            1600766733000000000,
        ]
    )
    event_time_offsets_1 = np.array([456, 743, 347, 345, 632])
    event_data_1 = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets_1,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_1_ids = np.array([0, 1, 2, 3])
    x_pixel_offset_1 = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset_1 = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset_1 = np.array([0.1, 0.2, 0.3, 0.4])

    distance = 57  # cm
    transformation = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array(distance),
        value_units="cm",
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(
            detector_1_ids,
            event_data_1,
            x_offsets=x_pixel_offset_1,
            y_offsets=y_pixel_offset_1,
            z_offsets=z_pixel_offset_1,
            offsets_unit="m",
            depends_on=transformation,
        )
    )

    loaded_data = load_function(builder)

    expected_pixel_positions = np.array(
        [x_pixel_offset_1, y_pixel_offset_1, z_pixel_offset_1]
    ).T
    assert np.allclose(
        get_meta(loaded_data)['base_position'].values, expected_pixel_positions
    )

    expected_transform = sc.spatial.translation(unit=sc.units.m, value=[0, 0, 0.57])

    assert np.allclose(
        get_meta(loaded_data)['position_transformations'].value.values,
        expected_transform.values,
    )

    assert np.allclose(
        (
            get_meta(loaded_data)["position_transformations"].value
            * get_meta(loaded_data)["base_position"]["detector_id", 0]
        ).values,
        [0.1, 0.1, 0.67],
    )

    assert np.allclose(
        loaded_data.coords["position"]["detector_id", 0].values, [0.1, 0.1, 0.67]
    )


def test_loads_pixel_positions_with_multiple_transformations(load_function: Callable):
    event_data_1 = EventData(
        event_id=np.array([0, 0, 0, 0, 0]),
        event_time_offset=(np.array([456, 743, 347, 345, 632])),
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )
    event_data_2 = EventData(
        event_id=np.array([1, 1, 1, 1, 1]),
        event_time_offset=(np.array([456, 743, 347, 345, 632])),
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    transformation1 = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array(12),
        value_units="cm",
    )
    transformation2 = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array(34),
        value_units="cm",
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(
            np.array([0]),
            event_data_1,
            x_offsets=np.array([0.1]),
            y_offsets=np.array([0.1]),
            z_offsets=np.array([0.1]),
            offsets_unit="m",
            depends_on=transformation1,
        )
    )

    builder.add_detector(
        Detector(
            np.array([1]),
            event_data_2,
            x_offsets=np.array([0.6]),
            y_offsets=np.array([0.6]),
            z_offsets=np.array([0.6]),
            offsets_unit="m",
            depends_on=transformation2,
        )
    )

    loaded_data = load_function(builder)

    assert np.allclose(
        (
            get_meta(loaded_data)["position_transformations"]["detector_id", 0].value
            * get_meta(loaded_data)["base_position"]["detector_id", 0]
        ).values,
        [0.1, 0.1, 0.1 + 0.12],
    )
    assert np.allclose(
        (
            get_meta(loaded_data)["position_transformations"]["detector_id", 1].value
            * get_meta(loaded_data)["base_position"]["detector_id", 1]
        ).values,
        [0.6, 0.6, 0.6 + 0.34],
    )


def test_links_to_event_data_group_are_ignored(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    builder.add_hard_link(Link("/entry/hard_link_to_events", "/entry/events_0"))
    builder.add_soft_link(Link("/entry/soft_link_to_events", "/entry/events_0"))

    loaded_data = load_function(builder)

    # The output Variable must contain the events from the added event
    # dataset with no duplicate data due to the links

    # Expect time of flight to match the values in the
    # event_time_offset dataset
    # May be reordered due to binning (hence np.sort)
    assert np.array_equal(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(event_time_offsets),
    )

    counts_on_detectors = loaded_data.bins.sum()
    # No detector_number dataset in file so expect detector_id to be
    # binned according to whatever detector_ids are present in event_id
    # dataset: 2 on det 1, 1 on det 2, 2 on det 3
    expected_counts = np.array([[2], [1], [2]])
    assert np.array_equal(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.array([1, 2, 3])
    assert np.array_equal(
        loaded_data.coords['detector_id'].values, expected_detector_ids
    )


def test_links_in_transformation_paths_are_followed(load_function: Callable):
    builder = NexusBuilder()
    distance = 13.6
    builder.add_component(Source("source"))
    builder.add_dataset_at_path(
        "/entry/transform",
        np.array(distance),
        {
            "vector": np.array([0, 0, 1]),
            "units": "m",
            "transformation_type": "translation",
            "depends_on": ".",
        },
    )
    builder.add_dataset_at_path("/entry/source/depends_on", "/entry/transform_link", {})
    builder.add_soft_link(Link("/entry/transform_link", "/entry/transform"))
    loaded_data = load_function(builder)

    assert np.allclose(loaded_data["source_position"].values, [0, 0, distance])
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_relative_links_in_transformation_paths_are_followed(load_function: Callable):
    builder = NexusBuilder()
    distance = 13.6
    builder.add_component(Source("source"))
    builder.add_dataset_at_path(
        "/entry/transform1",
        np.array(distance),
        {
            "vector": np.array([0, 0, 1]),
            "units": "m",
            "transformation_type": "translation",
            "depends_on": ".",
        },
    )
    builder.add_dataset_at_path(
        "/entry/transform2",
        np.array(distance),
        {
            "vector": np.array([0, 0, 1]),
            "units": "m",
            "transformation_type": "translation",
            "depends_on": "transform1",
        },
    )
    builder.add_dataset_at_path("/entry/source/depends_on", "/entry/transform2", {})
    loaded_data = load_function(builder)

    assert np.allclose(loaded_data["source_position"].values, [0, 0, 2 * distance])
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_linked_datasets_are_found(load_function: Callable):
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=np.array([456, 743, 347, 345, 632]),
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    replaced_ids = np.array([1, 1, 1, 2, 3])
    builder.add_dataset_at_path("/entry/ids", replaced_ids, {})
    replaced_tofs = np.array([273, 546, 573, 812, 932])
    builder.add_dataset_at_path("/entry/tofs", replaced_tofs, {})
    # Replace dataset in the NXevent_data with a link to the
    # replacement dataset
    builder.add_hard_link(Link("/entry/events_0/event_id", "/entry/ids"))
    builder.add_soft_link(Link("/entry/events_0/event_time_offset", "/entry/tofs"))

    loaded_data = load_function(builder)

    assert np.array_equal(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(replaced_tofs),
    )

    counts_on_detectors = loaded_data.bins.sum()
    expected_counts = np.array([[3], [1], [1]])
    assert np.array_equal(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.array([1, 2, 3])
    assert np.array_equal(
        loaded_data.coords['detector_id'].values, expected_detector_ids
    )


def test_loads_sample_ub_matrix(load_function: Callable):
    builder = NexusBuilder()
    builder.add_component(Sample("sample", ub_matrix=np.ones(shape=[3, 3])))
    loaded_data = load_function(builder)
    sample = loaded_data['sample'].value
    assert sc.identical(
        sample['ub_matrix'],
        sc.spatial.linear_transform(
            value=np.ones(shape=[3, 3]), unit=sc.units.angstrom**-1
        ),
    )
    assert "orientation_matrix" not in sample


def test_loads_sample_orientation_matrix(load_function: Callable):
    builder = NexusBuilder()
    builder.add_component(Sample("sample", orientation_matrix=np.ones(shape=[3, 3])))
    loaded_data = load_function(builder)
    sample = loaded_data['sample'].value
    assert sc.identical(
        sample["orientation_matrix"],
        sc.spatial.linear_transform(value=np.ones(shape=[3, 3]), unit=sc.units.one),
    )
    assert "ub_matrix" not in sample


def test_loads_multiple_sample_ub_matrix(load_function: Callable):
    builder = NexusBuilder()
    builder.add_component(Sample("sample1", ub_matrix=np.ones(shape=[3, 3])))
    builder.add_component(Sample("sample2", ub_matrix=np.identity(3)))
    builder.add_component(Sample("sample3"))  # No ub specified
    loaded_data = load_function(builder)
    assert sc.identical(
        loaded_data["sample1"].value["ub_matrix"],
        sc.spatial.linear_transform(
            value=np.ones(shape=[3, 3]), unit=sc.units.angstrom**-1
        ),
    )
    assert sc.identical(
        loaded_data["sample2"].value["ub_matrix"],
        sc.spatial.linear_transform(value=np.identity(3), unit=sc.units.angstrom**-1),
    )
    assert "ub_matrix" not in loaded_data['sample3'].value


def test_warning_but_no_error_for_unrecognised_log_unit(load_function: Callable):
    values = np.array([1.1, 2.2, 3.3])
    times = np.array([4.4, 5.5, 6.6])
    name = "test_log"
    builder = NexusBuilder()
    unknown_unit = "elephants"
    builder.add_log(Log(name, values, times, value_units=unknown_unit, time_units="s"))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert name in loaded_data


def test_start_and_end_times_appear_in_dataset_if_set(load_function: Callable):
    builder = NexusBuilder()
    builder.add_run_start_time("2001-01-01T00:00:00")
    builder.add_run_end_time("2002-02-02T00:00:00")

    loaded_data = load_function(builder)

    assert sc.identical(
        loaded_data["start_time"], sc.DataArray(sc.scalar("2001-01-01T00:00:00"))
    )
    assert sc.identical(
        loaded_data["end_time"], sc.DataArray(sc.scalar("2002-02-02T00:00:00"))
    )


@pytest.mark.parametrize(
    "log_start,scaling_factor",
    (
        ("2000-01-01T01:00:00", 1000),
        ("2000-01-01T00:00:00", 0.001),
        ("2010-01-01T00:00:00", None),
    ),
)
def test_load_log_times(log_start: str, scaling_factor: float, load_function: Callable):
    times = np.array([0.0, 10.0, 20.0, 30.0, 40.0, 50.0], dtype="float64")

    builder = NexusBuilder()
    builder.add_log(
        Log(
            name="test_log",
            value=np.zeros(shape=(len(times),)),
            time=times,
            start_time=log_start,
            scaling_factor=scaling_factor,
        )
    )

    loaded_data = load_function(builder)

    times_ns = sc.to_unit(
        sc.array(
            dims=["time"],
            values=times * (scaling_factor or 1.0),
            unit=sc.units.s,
            dtype=sc.DType.float64,
        ),
        sc.units.ns,
    ).astype(sc.DType.int64)

    expected = (
        sc.scalar(
            value=np.datetime64(log_start), unit=sc.units.ns, dtype=sc.DType.datetime64
        )
        + times_ns
    )

    actual = loaded_data["test_log"].values.coords['time']

    diffs = np.abs(actual.values - expected.values)

    # Allow 1ns difference for rounding errors between different routes
    assert all(diffs <= np.array(1).astype("timedelta64[ns]"))


def test_load_log_times_when_logs_do_not_have_start_time(load_function: Callable):
    # If an NXLog doesn't have a start time attribute then 1970-01-01 should be
    # assumed instead

    times = np.array([-10.0, 0.0, 10.0, 20.0, 30.0, 40.0, 50.0], dtype="float64")

    builder = NexusBuilder()
    builder.add_log(
        Log(
            name="test_log",
            value=np.zeros(shape=(len(times),)),
            time=times,
            start_time=None,
        )
    )

    loaded_data = load_function(builder)

    times_ns = sc.to_unit(
        sc.array(dims=["time"], values=times, unit=sc.units.s, dtype=sc.DType.float64),
        sc.units.ns,
    ).astype(sc.DType.int64)

    expected = (
        sc.scalar(
            value=np.datetime64("1970-01-01T00:00:00"),
            unit=sc.units.ns,
            dtype=sc.DType.datetime64,
        )
        + times_ns
    )

    actual = loaded_data["test_log"].values.coords['time']

    diffs = np.abs(actual.values - expected.values)

    # Allow 1ns difference for rounding errors between different routes
    assert all(diffs <= np.array(1).astype("timedelta64[ns]"))


@pytest.mark.parametrize("units", ("ps", "ns", "us", "ms", "s", "minute", "hour"))
def test_adjust_log_times_with_different_time_units(units, load_function: Callable):
    times = [1, 2, 3]

    builder = NexusBuilder()
    builder.add_log(
        Log(
            name="test_log",
            value=np.zeros(shape=(len(times),)),
            time=np.array(times, dtype="float64"),
            time_units=units,
            start_time="2000-01-01T00:00:00",
        ),
    )

    loaded_data = load_function(builder)

    times_ns = sc.to_unit(
        sc.array(dims=["time"], values=times, unit=units, dtype=sc.DType.int64),
        sc.units.ns,
    )

    expected = (
        sc.scalar(
            value=np.datetime64("2000-01-01T00:00:00"),
            unit=sc.units.ns,
            dtype=sc.DType.datetime64,
        )
        + times_ns
    )

    actual = loaded_data["test_log"].values.coords['time']

    diffs = np.abs(actual.values - expected.values)

    # Allow 1ns difference for rounding errors between different routes
    assert all(diffs <= np.array(1).astype("timedelta64[ns]"))


def test_nexus_file_with_invalid_nxlog_time_units_loads_dataset_as_non_datetime(
    load_function: Callable,
):
    builder = NexusBuilder()
    builder.add_log(
        Log(
            name="test_log_1",
            value=np.zeros(shape=(1,)),
            time=np.array([1.0]),
            time_units="m",  # Time in metres, should fail.
            start_time="1970-01-01T00:00:00",
        )
    )

    loaded_data = load_function(builder)
    assert loaded_data['test_log_1'].value.coords['time'].dtype == sc.DType('float64')
    assert loaded_data['test_log_1'].value.coords['time'].unit == 'm'


def test_nexus_file_with_invalid_log_start_date_uses_epoch(load_function: Callable):
    builder = NexusBuilder()
    builder.add_log(
        Log(
            name="test_log_1",
            value=np.zeros(shape=(1,)),
            time=np.array([1]),
            start_time="this_isnt_a_valid_log_start_time",
        )
    )
    builder.add_log(
        Log(
            name="test_log_2",
            value=np.zeros(shape=(1,)),
            time=np.array([1]),
            start_time="1970-01-01T00:00:00",
        )
    )

    loaded_data = load_function(builder)
    assert sc.identical(
        loaded_data['test_log_1'].value.coords['time'],
        sc.epoch(unit='s')
        + sc.array(dims=['time'], values=[1], unit='s', dtype='int64'),
    )


def test_extended_ascii_in_ascii_encoded_dataset(load_function: Callable):
    if load_function == load_from_json:
        pytest.skip("JSON serialiser can only serialize strings, not bytes.")

    builder = NexusBuilder()
    # When writing, if we use bytes h5py will write as ascii encoding
    # 0xb0 = degrees symbol in latin-1 encoding.
    builder.add_title(b"run at rot=90" + bytes([0xB0]))

    with pytest.warns(UserWarning, match="contains characters in extended ascii range"):
        loaded_data = load_function(builder)

        assert sc.identical(
            loaded_data["experiment_title"],
            sc.DataArray(data=sc.scalar("run at rot=90°")),
        )


@pytest.mark.parametrize("test_string", UTF8_TEST_STRINGS)
def test_utf8_encoded_dataset(load_function: Callable, test_string):
    builder = NexusBuilder()
    # When writing, if we use str h5py will write as utf8 encoding
    builder.add_title(test_string)

    loaded_data = load_function(builder)

    assert sc.identical(
        loaded_data["experiment_title"], sc.DataArray(data=sc.scalar(test_string))
    )


def test_extended_ascii_in_ascii_encoded_attribute(load_function: Callable):
    if load_function == load_from_json:
        pytest.skip("JSON serialiser can only serialize strings, not bytes.")

    builder = NexusBuilder()
    # When writing, if we use bytes h5py will write as ascii encoding
    # 0xb0 = degrees symbol in latin-1 encoding.
    builder.add_log(Log(name="testlog", value_units=bytes([0xB0]), value=np.array([0])))

    with pytest.warns(UserWarning, match="contains characters in extended ascii range"):
        loaded_data = load_function(builder)

        assert loaded_data["testlog"].data.values.unit == sc.units.deg


# Can't use UTF-8 test strings as above for this test as the units need to be valid.
# Just do a single test with degrees.
def test_utf8_encoded_attribute(load_function: Callable):
    builder = NexusBuilder()
    # When writing, if we use str h5py will write as utf8 encoding
    builder.add_log(Log(name="testlog", value_units="°", value=np.array([0])))

    loaded_data = load_function(builder)
    assert loaded_data["testlog"].data.values.unit == sc.units.deg


def test_load_nexus_adds_single_tof_bin(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632], dtype="float64")
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array(
            [
                1600766730000000000,
                1600766731000000000,
                1600766732000000000,
                1600766733000000000,
            ]
        ),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)

    loaded_data = load_function(builder)

    # Size 2 for each of the two bin edges around a single bin
    assert loaded_data.coords["tof"].shape == (2,)

    # Assert bin edges correspond to smallest and largest+1 time-of-flights
    # in data.
    assert sc.identical(
        loaded_data.coords["tof"]["tof", 0],
        sc.scalar(value=np.min(event_time_offsets), unit=sc.units.ns),
    )
    assert sc.identical(
        loaded_data.coords["tof"]["tof", 1],
        sc.scalar(
            value=np.nextafter(np.max(event_time_offsets), float("inf")),
            unit=sc.units.ns,
        ),
    )


def test_nexus_file_with_choppers(load_function: Callable):
    builder = NexusBuilder()
    builder.add_instrument("dummy")
    builder.add_chopper(
        Chopper(
            "chopper_1",
            distance=10.0,
            rotation_speed=60.0,
            rotation_units="Hz",
            distance_units="m",
        )
    )
    loaded_data = load_function(builder)
    chopper = loaded_data['chopper_1'].value
    assert sc.identical(chopper["rotation_speed"], 60.0 * sc.Unit("Hz"))
    assert sc.identical(chopper["distance"], 10.0 * sc.Unit("m"))


def test_nexus_file_with_two_choppers(load_function: Callable):
    builder = NexusBuilder()
    builder.add_instrument("dummy")
    builder.add_chopper(
        Chopper(
            "chopper_1",
            distance=11.0 * 1000,
            rotation_speed=65.0 / 1000,
            rotation_units="MHz",
            distance_units="mm",
        )
    )
    builder.add_chopper(
        Chopper(
            "chopper_2",
            distance=10.0,
            rotation_speed=60.0,
            rotation_units="Hz",
            distance_units="m",
        )
    )
    loaded_data = load_function(builder)
    chopper1 = loaded_data['chopper_1'].value
    chopper2 = loaded_data['chopper_2'].value
    assert sc.identical(chopper1["rotation_speed"], (65.0 / 1000) * sc.Unit("MHz"))
    assert sc.identical(chopper1["distance"], (11.0 * 1000) * sc.Unit("mm"))
    assert sc.identical(chopper2["rotation_speed"], 60.0 * sc.Unit("Hz"))
    assert sc.identical(chopper2["distance"], 10.0 * sc.Unit("m"))


def test_load_monitors(load_function: Callable):
    builder = NexusBuilder()

    # Monitor with data
    builder.add_monitor(
        Monitor(
            name="monitor1",
            data=np.ones(shape=(2, 4, 6), dtype="float64"),
            axes=[
                ("event_index", np.arange(3, 5, dtype="float64")),
                ("period_index", np.arange(3, 7, dtype="float64")),
                ("time_of_flight", np.arange(3, 9, dtype="float64")),
            ],
        )
    )

    # Monitor with only one "axis" and only one data item
    builder.add_monitor(
        Monitor(
            "monitor2", data=np.array([1.0]), axes=[("time_of_flight", np.array([1.0]))]
        )
    )

    assert sc.identical(
        load_function(builder)["monitor1"].data.values,
        sc.DataArray(
            data=sc.ones(
                dims=["event_index", "period_index", "time_of_flight"],
                shape=(2, 4, 6),
                dtype=sc.DType.float64,
            ),
            coords={
                "event_index": sc.Variable(
                    dims=["event_index"], values=np.arange(3, 5, dtype="float64")
                ),
                "period_index": sc.Variable(
                    dims=["period_index"], values=np.arange(3, 7, dtype="float64")
                ),
                "time_of_flight": sc.Variable(
                    dims=["time_of_flight"], values=np.arange(3, 9, dtype="float64")
                ),
            },
        ),
    )

    assert sc.identical(
        load_function(builder)["monitor2"].data.values,
        sc.DataArray(
            data=sc.ones(dims=["time_of_flight"], shape=(1,)),
            coords={
                "time_of_flight": sc.Variable(
                    dims=["time_of_flight"], values=np.array([1.0])
                ),
            },
        ),
    )


def test_load_monitors_with_event_mode_data(load_function: Callable):
    builder = NexusBuilder()

    # Monitor with data
    builder.add_monitor(
        Monitor(
            name="monitor1",
            data=np.ones(shape=(2, 3, 4), dtype="float64"),
            axes=[
                ("event_index", np.array([0, 5], dtype="int64")),
                ("period_index", np.array([2, 3, 4], dtype="int64")),
                ("time_of_flight", np.array([3, 4, 5, 6], dtype="float64")),
            ],
            events=EventData(
                event_id=np.array([0, 0, 0, 0, 0], dtype="int64"),
                event_time_offset=np.array([1, 2, 3, 4, 5], dtype="float64"),
                event_time_zero=np.array([0, 1], dtype="float64"),
                event_index=np.array([0, 5], dtype="int64"),
            ),
        )
    )

    # Monitor with only one "axis" and only one data item
    builder.add_monitor(
        Monitor(
            "monitor2",
            data=np.array([1.0], dtype="float64"),
            axes=[("time_of_flight", np.array([1.0], dtype="float64"))],
            events=EventData(
                event_id=np.array([1, 1, 1, 1, 1], dtype="int64"),
                event_time_offset=np.array([6, 7, 8, 9, 10], dtype="float64"),
                event_time_zero=np.array([0, 1], dtype="float64"),
                event_index=np.array([0, 5], dtype="int64"),
            ),
        )
    )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message='Failed to load', category=UserWarning
        )
        warnings.filterwarnings("ignore", message='Skipped load', category=UserWarning)
        mon_1_events = load_function(builder)["monitor1"].data.values
        mon_2_events = load_function(builder)["monitor2"].data.values

    # Monitor 1 is too ambiguous for scippnexus to handle it
    assert isinstance(mon_1_events, sc.DataGroup)
    assert sc.identical(
        mon_1_events["event_time_offset"],
        sc.array(
            dims=["event"],
            values=[1, 2, 3, 4, 5],
            unit=sc.units.ns,
            dtype=sc.DType.float64,
        ),
    )
    assert sc.identical(
        mon_2_events.values[0].coords["tof"],
        sc.array(
            dims=["event"],
            values=[6, 7, 8, 9, 10],
            unit=sc.units.ns,
            dtype=sc.DType.float64,
        ),
    )


def test_load_monitor_with_transformation(load_function: Callable):
    builder = NexusBuilder()

    transformation = Transformation(
        TransformationType.TRANSLATION,
        vector=np.array([0, 0, 1]),
        value=np.array(6.5),
        value_units='m',
        offset=[1, 2, 3],
        offset_unit='m',
    )

    # Monitor with data and transformation
    builder.add_monitor(
        Monitor(
            name="monitor1",
            data=np.ones(shape=(2, 4, 6), dtype="float64"),
            axes=[
                ("event_index", np.arange(3, 5, dtype="float64")),
                ("period_index", np.arange(3, 7, dtype="float64")),
                ("time_of_flight", np.arange(3, 9, dtype="float64")),
            ],
            depends_on=transformation,
        )
    )

    loaded = load_function(builder)["monitor1"].data.values

    assert sc.identical(
        loaded.coords["position"], sc.vector(value=[1.0, 2.0, 9.5], unit="m")
    )


def test_load_nexus_file_containing_empty_arrays(load_function: Callable):
    event_data = EventData(
        event_id=np.array([], dtype='int32'),
        event_time_offset=np.array([]),
        event_time_zero=np.array([]),
        event_index=np.array([]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    builder.add_log(Log("test_log", np.array([0, 1, 2]), np.array([4, 5, 6])))

    # Empty datasets should not stop other data (e.g. metadata) from being loaded.
    loaded_data = load_function(builder)
    assert "test_log" in get_attrs(loaded_data)
    assert all(get_attrs(loaded_data)["test_log"].value.values == np.array([0, 1, 2]))



import scippneutron as scn


class LoadNexus:
    def setup(self):
        self.file_path = scn.data.get_path('PG3_4844_event.nxs')

    def time_load_event_nexus(self):
        scn.load_nexus(self.file_path)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass


@dataclass(frozen=True)
class StopTimeUpdate:
    stop_time_ms: int  # milliseconds from unix epoch
